/* AlignmentViewer.java
 *
 * created: Mon Jul 12 1999
 *
 * This file is part of Artemis
 * 
 * Copyright (C) 1999  Genome Research Limited
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /nfs/disk222/yeastpub/Repository/powmap/diana/components/AlignmentViewer.java,v 1.3 2000/03/09 11:46:02 kmr Exp $
 */

package diana.components;

import diana.*;

import java.awt.*;
import java.awt.event.*;

/**
 *  This component shows an alignment of two sequences using the data from a
 *  ComparisonData object.
 *
 *  @author Kim Rutherford
 *  @version $Id: AlignmentViewer.java,v 1.3 2000/03/09 11:46:02 kmr Exp $
 **/

public class AlignmentViewer extends Panel {
  /**
   *  Create a new AlignmentViewer for the given entries
   **/
  public AlignmentViewer (final EntryGroup first_entry_group,
                          final EntryGroup second_entry_group,
                          final ComparisonData comparison_data) {
    this.first_entry = first_entry;
    this.second_entry = second_entry;
    this.comparison_data = comparison_data;

    setLayout(new BorderLayout ()); 
    
    createCanvas ();

    add (new Scrollbar (), "East");
  }

  /**
   *  This method tells this AlignmentViewer component where the top sequence
   *  is now.
   **/
  public void setTopSequencePosition (final DisplayAdjustmentEvent event) {
    last_top_event = event;
    repaintCanvas (); 
  }

  /**
   *  This method tells this AlignmentViewer component where the bottom
   *  sequence is now.
   **/
  public void setBottomSequencePosition (final DisplayAdjustmentEvent event) {
    last_bottom_event = event;
    repaintCanvas (); 
  }
  
  /**
   *  Create the canvas object and add it.
   **/
  private void createCanvas () {
    canvas = new Canvas ()
      {
      /**
       *  Set the offscreen buffer to null as part of invalidation.
       **/
      public void invalidate() {
        super.invalidate();
        offscreen = null;
      }

      /**
       *  Override update to *not* erase the background before painting
       */
      public void update(final Graphics g) {
        paint(g);
      }

      /**
       *  Paint the canvas.
       */
      public void paint(final Graphics g) {
        paintCanvas (g);
      }

    };

    add (canvas, "Center");

    canvas.addMouseListener (new MouseAdapter () {
      /**
       *  Listen for mouse press events so that we can do popup menus and
       *  selection.
       **/
      public void mousePressed (final MouseEvent event) {
        //        System.out.println ("mousePressed ()");

      }
    });
  }

  /**
   *  Call repaint () on the canvas.
   **/
  private void repaintCanvas () {
    canvas.repaint ();
  }

  /**
   *  The main paint function for the canvas.  An off screen image used for
   *  double buffering when drawing the canvas.
   *  @param g The Graphics object of the canvas.
   **/
  private void paintCanvas (final Graphics g) {
    final int canvas_width = canvas.getSize ().width;
    final int canvas_height = canvas.getSize ().height;

    if (canvas_height <= 0 || canvas_width <= 0) {
      // there is no point drawing into a zero width canvas
      return;
    }

    if (offscreen == null) {
      offscreen = createImage (canvas_width,
                               canvas_height);
    }

    Graphics og = offscreen.getGraphics ();
    og.setClip (0, 0, canvas_width, canvas_height);

    fillBackground (og);

    if (last_top_event != null && last_bottom_event != null) {
      drawAlignments (og);
    }
      
    g.drawImage (offscreen, 0, 0, null);
    og.dispose ();
  }
  
  /**
   *  Draw the background colour of the frames.
   **/
  private void fillBackground (final Graphics g) {
    g.setColor (Color.white);

    g.fillRect (0, 0, canvas.getSize ().width, canvas.getSize ().height);
  }

  /**
   *  Draw the alignment into the given Graphics object.
   **/
  private void drawAlignments (final Graphics g) {        
    final int canvas_height = canvas.getSize ().height;
    final int canvas_width = canvas.getSize ().width;

    final AlignMatch [] alignment_data_array =
      comparison_data.getAlignments ();

    for (int i = 0 ; i < alignment_data_array.length ; ++i) {
      final AlignMatch this_match = alignment_data_array [i];

      final int top_start_x =
        getScreenPosition (canvas_width, last_top_event,
                           this_match.getFirstSequenceStart ());
      final int top_end_x =
        getScreenPosition (canvas_width, last_top_event,
                           this_match.getFirstSequenceEnd ());

      final int bottom_start_x =
        getScreenPosition (canvas_width, last_bottom_event,
                           this_match.getSecondSequenceStart ());
      final int bottom_end_x =
        getScreenPosition (canvas_width, last_bottom_event,
                           this_match.getSecondSequenceEnd ());

      if (top_end_x < 0 && bottom_end_x < 0 ||
          top_start_x >= canvas_width && bottom_start_x >= canvas_width) {
        continue;
      }
      
      final int [] x_coords = new int [4];
      final int [] y_coords = new int [4];
      
      x_coords[0] = top_start_x;
      y_coords[0] = 0;
      x_coords[1] = bottom_start_x;
      y_coords[1] = canvas_height - 1;
      x_coords[2] = bottom_end_x;
      y_coords[2] = canvas_height - 1;
      x_coords[3] = top_end_x;
      y_coords[3] = 0;
      
      g.setColor (Color.red);

      g.fillPolygon (x_coords, y_coords, x_coords.length);

      g.setColor (Color.black);

      g.drawLine (top_start_x, 0, bottom_start_x, canvas_height);
      g.drawLine (top_end_x, 0, bottom_end_x, canvas_height);
    }
  }

  /**
   *  Convert a base position into a screen x coordinate.
   **/
  private int getScreenPosition (final int canvas_width,
                                 final DisplayAdjustmentEvent event,
                                 final int base_position) {
    // this is the base that is at the left of the screen
    final int screen_start_base = event.getStart ();
    
    final double base_pos_float =
      event.getBaseWidth () * (base_position - screen_start_base);

    if (base_pos_float > 30000) {
      return 30000;
    } else {
      if (base_pos_float < -30000) {
        return -30000;
      } else {
        return (int) base_pos_float;
      }
    }
  }

  /**
   *  One of the two Entry objects that we are comparing.
   **/
  final Entry first_entry;

  /**
   *  One of the two Entry objects that we are comparing.
   **/
  final Entry second_entry;

  /**
   *  The comparison data that will be displayed in this component.
   **/
  final ComparisonData comparison_data;

  /**
   *  The drawing area for this component.
   **/
  private Canvas canvas = null;

  /**
   *  Off screen image used for double buffering when drawing the canvas.
   **/
  private Image offscreen;

  /**
   *  This is the last DisplayAdjustmentEvent reference that was passed to
   *  setTopSeqeuencePosition ().
   **/
  private DisplayAdjustmentEvent last_top_event; 

  /**
   *  This is the last DisplayAdjustmentEvent reference that was passed to
   *  setBottomSeqeuencePosition ().
   **/
  private DisplayAdjustmentEvent last_bottom_event; 
}

