/* Comparator.java
 *
 * created: Mon Jul 12 1999,2000
 *
 * This file is part of Artemis
 *
 * Copyright (C) 1999  Genome Research Limited
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /nfs/disk222/yeastpub/Repository/powmap/diana/components/Comparator.java,v 1.4 2000/03/09 15:27:21 kmr Exp $
 */

package diana.components;

import diana.*;
import diana.sequence.*;

import uk.ac.sanger.pathogens.OutOfRangeException;

import java.awt.*;
import java.awt.event.*;
import java.util.Vector;
import java.util.EventObject;

/**
 *  This Frame component contains a AlignmentViewer component, two
 *  FeatureDisplay components and the event listeners that keep them
 *  synchronized.
 *
 *  @author Kim Rutherford
 *  @version $Id: Comparator.java,v 1.4 2000/03/09 15:27:21 kmr Exp $
 **/

public class Comparator extends Frame
    implements GotoEventSource, FeatureDisplayOwner {
  /**
   *  Create a new Comparator component using the given data.
   **/
  public Comparator (final EntryGroup first_entry_group,
                     final EntryGroup second_entry_group,
                     final ComparisonData comparison_data) {
    this.first_entry_group = first_entry_group;
    this.second_entry_group = second_entry_group;
    this.comparison_data = comparison_data;

    button_panel = new Panel ();

    final FlowLayout flow_layout = new FlowLayout (FlowLayout.LEFT);

    button_panel.setLayout (flow_layout);

    lock_checkbox = new Checkbox ("lock", true);

    button_panel.add (lock_checkbox);

    final Selection first_selection =
      new Selection (getToolkit().getSystemClipboard());
    final Selection second_selection =
      new Selection (getToolkit().getSystemClipboard());

    top_entry_display =
      new FeatureDisplay (first_entry_group, first_selection,
                          this, this, true);
    alignment_viewer =
      new AlignmentViewer (first_entry_group, second_entry_group,
                           comparison_data);
    bottom_entry_display =
      new FeatureDisplay (second_entry_group, second_selection,
                          this, this, false);

    top_entry_display.setShowLabels (false);
    bottom_entry_display.setShowLabels (false);

    first_entry_group.addFeatureChangeListener (first_selection);
    second_entry_group.addEntryChangeListener (second_selection);

    addDisplayListeners (top_entry_display, bottom_entry_display);

    setLayout(new BorderLayout ());

    setTitle ("Dual Artemis Display");

    if (first_entry_group.getDefaultEntry () != null &&
        second_entry_group.getDefaultEntry () != null) {
      final String first_name =
        first_entry_group.getDefaultEntry ().getName ();
      final String second_name =
        second_entry_group.getDefaultEntry ().getName ();

      if (first_name != null && second_name != null) {
        setTitle ("Dual Artemis Display: " + first_name + " vs " +
                  second_name);
      }
    }

    final Font default_font = getDefaultFont ();

    setFont (default_font);

    addWindowListener (new WindowAdapter () {
      public void windowClosing (WindowEvent event) {
        // XXX
      }
    });

    GridBagLayout gridbag = new GridBagLayout();
    setLayout (gridbag);

    GridBagConstraints c = new GridBagConstraints();

    c.gridwidth  = GridBagConstraints.REMAINDER;
    c.fill       = GridBagConstraints.HORIZONTAL;
    c.anchor     = GridBagConstraints.NORTH;
    c.gridheight = 1;
    c.weightx    = 1;
    c.weighty    = 0;

    gridbag.setConstraints (button_panel, c);
    add (button_panel);

    gridbag.setConstraints (top_entry_display, c);
    add (top_entry_display);

    c.fill = GridBagConstraints.BOTH;
    c.weighty = 1;
    gridbag.setConstraints (alignment_viewer, c);
    add (alignment_viewer);

    c.weighty = 0;
    gridbag.setConstraints (bottom_entry_display, c);
    add (bottom_entry_display);

    packme ();

    setSize (900,480);

    final Dimension screen = Toolkit.getDefaultToolkit().getScreenSize();
    setLocation (new Point ((screen.width - getSize ().width) / 2,
                            (screen.height - getSize ().height) / 2));
  }

  /**
   *  This method will call pack () and then move the Frame to the centre of
   *  the screen.  (Implementation of the FeatureDisplayOwner interface).
   **/
  public void packme () {
    pack ();

    final Dimension screen = Toolkit.getDefaultToolkit().getScreenSize();

    final int x_position = (screen.width - getSize ().width) / 2;
    int y_position = (screen.height - getSize ().height) / 2;

    if (y_position < 10) {
      y_position = 10;
    }

    setLocation (new Point (x_position, y_position));
  }

  /**
   *  This method sends an GotoEvent to all the GotoEvent listeners that will
   *  make the given base visible.
   *  @param base_marker The base to make visible.
   **/
  public void gotoBase (final Marker base_marker) {
    final GotoEvent new_event = new GotoEvent (this, base_marker);

    sendGotoEvent (new_event);
  }

  /**
   *  Return an object that implements the GotoEventSource interface, and is
   *  the controlling object for Goto events associated with this object.
   **/
  public GotoEventSource getGotoEventSource () {
    return this;
  }

  /**
   *  Send the given event to all the GotoListeners.
   **/
  public void sendGotoEvent (final GotoEvent goto_event) {
    // don't send if there is nowhere to go to
    if (goto_event.getMarker () != null) {
      fireAction (goto_listener_list, goto_event);
    }
  }

  /**
   *  This method sends an GotoEvent to all the GotoEvent listeners that will
   *  make the given base visible.  This just calls makeBaseVisible ().
   *  @param destination_base The base to make visible.
   *  @return The reference of a MarkerRange object for the given base or null
   *    if the call fails.  It will fail only if the destination_base is less
   *    than 1 or greater than the length of the sequence.
   **/
  public MarkerRange gotoBase (final int destination_base) {
//      try {
//        final Strand forward_strand =
//          getEntryGroup ().getBases ().getForwardStrand ();
//        final Marker destination_marker =
//          forward_strand.makeMarker (destination_base);
//        gotoBase (destination_marker);
//        return new MarkerRange (destination_marker.getStrand (),
//                                destination_base, destination_base);
//      } catch (OutOfRangeException e) {
      return null;
//      }
  }


  /**
   *  This method finds the given amino acid seqeunce, sets the selection to
   *  the matching bases and then sends a GotoEvent to all the GotoEvent
   *  listeners that will make the first base of the match visible.
   *  @param sequence This is the pattern to search for.
   *  @param start_at_first_base If true or if there is nothing in the
   *    Selection then the search will start at base 1, otherwise the search
   *    will start at the Selection.
   *  @return The range that matches the pattern or null if there is no match.
   **/
  public MarkerRange gotoAminoAcidSequence (final AminoAcidSequence sequence,
                                            boolean start_at_first_base) {
    return null;
  }

  /**
   *  This method finds the given pattern and then sends an GotoEvent to all
   *  the GotoEvent listeners that will make the first base of the match
   *  visible.
   *  @param pattern This is the pattern to search for.
   *  @param start_at_first_base If true or if there is nothing in the
   *    Selection then the search will start at base 1, otherwise the search
   *    will start at the Selection.
   *  @return The range that matches the pattern or null if there is no match.
   **/
  public MarkerRange gotoBasePattern (final BasePattern pattern,
                                      boolean start_at_first_base) {
    // XXX
    return null;

//      // if start_at_first_base is false we want to start the search at the
//      // selection
//      final Marker selection_base;

//      final Marker first_selection_base =
//        getSelection ().getFirstBaseOfSelection ();

//      final Marker last_selection_base =
//        getSelection ().getLastBaseOfSelection ();

//      // these checks are currently unnecessary, because if first_selection_base
//      // is null then last_selection_base will be too.  This is future proofing
//      if (first_selection_base == null || last_selection_base == null) {
//        if (first_selection_base == null) {
//          selection_base = last_selection_base;
//        } else {
//          selection_base = first_selection_base;
//        }
//      } else {
//        selection_base =
//          Marker.getRawLowest (first_selection_base, last_selection_base);
//      }

//      final Marker start_position;

//      if (start_at_first_base || selection_base == null) {
//        // null means start the search from the beginning
//        start_position = null;
//      } else {
//        start_position = selection_base;
//      }

//      final MarkerRange match_range =
//        pattern.findMatch (getEntryGroup ().getBases (),
//                           start_position,
//                           getEntryGroup ().getSequenceLength ());

//      if (match_range == null) {
//        return null;
//      } else {
//        gotoBase (match_range.getStart ());
//        return match_range;
//      }
  }


  /**
   *  This method sends a GotoEvent to all the GotoEvent listeners that will
   *  make the first base of the sequence visible.
   **/
  public void gotoFirstBase () {
    gotoBase (1);
  }

  /**
   *  This method sends a GotoEvent to all the GotoEvent listeners that will
   *  make the last base of the sequence visible.
   **/
  public void gotoLastBase () {
//      gotoBase (getEntryGroup ().getSequenceLength ());
  }

  /**
   *  Adds the specified event listener to receive Goto events from this
   *  object.
   *  @param l the GotoEvent listener.
   **/
  public void addGotoListener (final GotoListener l) {
    goto_listener_list.addElement (l);
  }

  /**
   *  Removes the specified event listener so that it no longer receives Goto
   *  events from this object.
   *  @param l the GotoEvent listener.
   **/
  public void removeGotoListener (final GotoListener l) {
    goto_listener_list.removeElement (l);
  }

  /**
   *  Wire-up the two FeatureDisplay objects with DisplayAdjustmentListeners.
   **/
  private void addDisplayListeners (final FeatureDisplay top_display,
                                    final FeatureDisplay bottom_display) {

    top_listener = new DisplayAdjustmentListener () {
      public void displayAdjustmentValueChanged (DisplayAdjustmentEvent e) {
        top_display.removeDisplayAdjustmentListener (top_listener);
        bottom_display.removeDisplayAdjustmentListener (bottom_listener);

        final int current_start = bottom_display.getFirstVisibleForwardBase ();

        bottom_display.setScaleFactor (e.getScaleFactor ());

        if (lock_checkbox.getState ()) {
          final int difference;

          if (last_top_event == null) {
            difference = 0;
          } else {
            difference = e.getStart () - last_top_event.getStart ();
          }

          final int new_start = current_start + difference;

          bottom_display.setFirstBase (new_start);
        }

        top_display.addDisplayAdjustmentListener (top_listener);
        bottom_display.addDisplayAdjustmentListener (bottom_listener);

        last_top_event = e;
      }
    };

    bottom_listener = new DisplayAdjustmentListener () {
      public void displayAdjustmentValueChanged (DisplayAdjustmentEvent e) {
        top_display.removeDisplayAdjustmentListener (top_listener);
        bottom_display.removeDisplayAdjustmentListener (bottom_listener);

        final int current_start = top_display.getFirstVisibleForwardBase ();

        top_display.setScaleFactor (e.getScaleFactor ());

        if (lock_checkbox.getState ()) {
          final int difference;

          if (last_bottom_event == null) {
            difference = 0;
          } else {
            difference = e.getStart () - last_bottom_event.getStart ();
          }

          final int new_start = current_start + difference;

          top_display.setFirstBase (new_start);
        }

        top_display.addDisplayAdjustmentListener (top_listener);
        bottom_display.addDisplayAdjustmentListener (bottom_listener);

        last_bottom_event = e;
      }
    };

    top_display.addDisplayAdjustmentListener (top_listener);
    bottom_display.addDisplayAdjustmentListener (bottom_listener);

    final DisplayAdjustmentListener top_align_listener =
      new DisplayAdjustmentListener () {
        public void displayAdjustmentValueChanged (DisplayAdjustmentEvent e) {
          alignment_viewer.setTopSequencePosition (e);
        }
      };

    top_display.addDisplayAdjustmentListener (top_align_listener);

    final DisplayAdjustmentListener bottom_align_listener =
      new DisplayAdjustmentListener () {
        public void displayAdjustmentValueChanged (DisplayAdjustmentEvent e) {
          alignment_viewer.setBottomSequencePosition (e);
        }
      };

    bottom_display.addDisplayAdjustmentListener (bottom_align_listener);
  }

  /**
   *  Send an event to those object listening for it.
   *  @param listeners A Vector of the objects that the event should be sent
   *    to.
   *  @param event The event to send
   **/
  private void fireAction (final Vector listeners, final EventObject event) {
    final Vector targets;
    // copied from a book - synchronising the whole method might cause a
    // deadlock
    synchronized (this) {
      targets = (Vector) listeners.clone ();
    }

    for (int i = 0 ; i < targets.size () ; ++i) {
      GotoListener target = (GotoListener) targets.elementAt (i);

      if (event instanceof GotoEvent) {
        final GotoListener goto_listener = (GotoListener) target;
        goto_listener.performGoto ((GotoEvent) event);
      } else {
        throw new Error ("EntryEdit.fireAction () - unknown event");
      }
    }
  }

  /**
   *  Return the current default font (from Diana.options).
   **/
  private Font getDefaultFont () {
    return Options.getOptions ().getFont ();
  }

  /**
   *  A vector of those objects listening for selection change events.
   **/
  final private Vector goto_listener_list = new Vector ();

  /**
   *  One of the two Entry objects that we are comparing.
   **/
  final private EntryGroup first_entry_group;

  /**
   *  One of the two Entry objects that we are comparing.
   **/
  final private EntryGroup second_entry_group;

  /**
   *  The comparison data that will be displayed in the AlignmentViewer
   *  component.
   **/
  final private ComparisonData comparison_data;

  final private Panel button_panel;
  final private Checkbox lock_checkbox;
  final private FeatureDisplay top_entry_display;
  final private FeatureDisplay bottom_entry_display;
  final private AlignmentViewer alignment_viewer;

  private DisplayAdjustmentListener top_listener = null;
  private DisplayAdjustmentListener bottom_listener = null;

  private DisplayAdjustmentEvent last_top_event;
  private DisplayAdjustmentEvent last_bottom_event;
}
