/* EntryChooser.java
 *
 * created: Wed Dec 23 1998
 *
 * This file is part of Artemis
 * 
 * Copyright (C) 1998,1999,2000  Genome Research Limited
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /nfs/disk222/yeastpub/Repository/powmap/diana/components/EntryChooser.java,v 1.9 2000/01/24 17:04:18 kmr Exp $
 **/

package diana.components;

import diana.*;
import java.awt.*;
import java.awt.event.*;
import java.util.Vector;

/**
 *  This component allows the user to select an Entry from a list.
 *
 *  @author Kim Rutherford
 *  @version $Id: EntryChooser.java,v 1.9 2000/01/24 17:04:18 kmr Exp $
 **/

public class EntryChooser extends Frame
    implements EntryGroupChangeListener, EntryChangeListener {
  /**
   *  Create a new EntryChooser component.
   *  @param entry_group The EntryGroup to choose an Entry from.
   *  @param message A message to prompt the user with.
   **/
  public EntryChooser (final EntryGroup entry_group,
                       final String message) {
    this.entry_group = entry_group;
    this.message = message;

    label = new Label (message);
    
    setLayout (new GridLayout (0, 1));
    
    refreshButtons ();
    
    addWindowListener (new WindowAdapter () {
      public void windowClosing (WindowEvent event) {
        dispose ();
      }
    });
    
    entry_group.addEntryGroupChangeListener (this);
    entry_group.addEntryChangeListener (this);

    packme ();
  }

  /**
   *  Remove this component as a entry group and entry change listener.
   **/
  private void stopListening () {
    entry_group.removeEntryGroupChangeListener (this);
    entry_group.removeEntryChangeListener (this);
  }

  /**
   *  Tell this component to stop listening for EntryGroupChange and
   *  EntryChange events then dispose () of this component.
   **/
  public void dispose () {
    stopListening ();
    super.dispose ();
  }

  /**
   *  Implementation of the EntryGroupChangeListener interface.  We listen to
   *  EntryGroupChange events so that we can update the display if entries
   *  are added or deleted.
   **/
  public void entryGroupChanged (EntryGroupChangeEvent event) {
//      System.out.println ("in FeatureList entry group change event type: " +
//                          event.getType () + " for entry " +
//                          event.getEntry ());

    switch (event.getType ()) {
    case EntryGroupChangeEvent.ENTRY_ADDED:
    case EntryGroupChangeEvent.ENTRY_DELETED:
      refreshButtons ();
      break;
    case EntryGroupChangeEvent.ENTRY_INACTIVE:
    case EntryGroupChangeEvent.ENTRY_ACTIVE:
      // ignore these events
      break;
    }
  }

  /**
   *  Implementation of the EntryChangeListener interface.
   **/
  public void entryChanged (final EntryChangeEvent event) {
//     System.out.println ("in EntryGroup: entry change event type: " +
//                         event.getType () + " for entry " +
//                         event.getSource () + " for feature " +
//                         event.getFeature ());

    if (event.getType () == EntryChangeEvent.NAME_CHANGED) {
      refreshButtons ();
    }
  }

  /**
   *  Add the given object as a listen for EntryChooser events from this
   *  EntryChooser.
   **/
  public void addEntryChooserListener (EntryChooserListener l) {
    entry_chooser_listeners.addElement (l);
  }

  /**
   *  Update the Buttons to the reflect the current contents of the
   *  EntryGroup. 
   **/
  private void refreshButtons () {
    setVisible (false);

    removeAll ();

    entry_components = new Vector ();

    add (label, "North");

    for (int i = 0 ; i < entry_group.size () ; ++i) {
      add (entry_group.elementAt (i));
    }

    add (button_panel, "South");
    
    close_button.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent e) {
        dispose ();
      }
    });

    button_panel.removeAll ();
    button_panel.add (close_button);
    
    packme ();

    setVisible (true);
  }

  /**
   *  Pack and position the Frame.
   **/
  private void packme () {
    pack ();

    final Dimension screen = Toolkit.getDefaultToolkit().getScreenSize();
    setLocation (new Point ((screen.width - getSize ().width) / 2,
                            (screen.height - getSize ().height) / 2));
  }

  /**
   *  Add a Label or Checkbox for the given Entry to this component.
   **/
  private void add (Entry entry) {
    final String entry_name;

    if (entry.getName () == null) {
      entry_name = "no name";
    } else {
      entry_name = entry.getName ();
    }

    final Button new_button = new Button (entry_name);

    new_button.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        final int button_index =
          entry_components.indexOf (event.getSource ());

        final Entry the_entry = entry_group.elementAt (button_index);

        // sent the event to all the listeners
        for (int i = 0 ; i < entry_chooser_listeners.size () ; ++i) {
          final EntryChooserListener listener =
            ((EntryChooserListener) entry_chooser_listeners.elementAt (i));

          final EntryChooserEvent new_event =
            new EntryChooserEvent (EntryChooser.this, the_entry);

          listener.entryChoosen (new_event);
        }
      }
    });

    entry_components.addElement (new_button);
    add (new_button);
    pack ();
  }

  /**
   *  This is the EntryGroup object that we are displaying.
   **/
  private EntryGroup entry_group;

  /**
   *  A message to prompt the user with.  This is the message String that is
   *  passed to the constructor.
   **/
  private String message;

  /**
   *  A vector containing one Button for each Entry in the EntryGroup object.
   **/
  private Vector entry_components = new Vector ();

  /**
   *  This contains the objects that are listening for EntryChooser events
   *  from this EntryChooser.
   **/
  private Vector entry_chooser_listeners = new Vector ();

  /**
   *  A label containing the user message (from the constructor).
   **/
  private final Label label;

  /**
   *  The Panel containing the close button. 
   **/
  private final Panel button_panel = new Panel ();

  /**
   *  The close button.
   **/
  private final Button close_button = new Button ("Close");
}
