/* EntryEdit.java
 *
 * created: Fri Oct  9 1998
 *
 * This file is part of Artemis
 *
 * Copyright (C) 1998,1999,2000  Genome Research Limited
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /nfs/disk222/yeastpub/Repository/powmap/diana/components/EntryEdit.java,v 1.127 2000/03/27 12:53:38 kmr Exp $
 */

package diana.components;

import Diana;

import diana.*;
import diana.plot.*;
import diana.sequence.*;

import uk.ac.sanger.pathogens.*;
import uk.ac.sanger.pathogens.embl.DocumentEntryFactory;
import uk.ac.sanger.pathogens.embl.EntryInformationException;
import uk.ac.sanger.pathogens.embl.InvalidKeyException;

import java.awt.*;
import java.awt.event.*;
import java.util.Vector;
import java.util.EventObject;
import java.io.IOException;
import java.io.File;
import java.io.FileReader;

/**
 *  Each object of this class is used to edit an EntryGroup object.
 *
 *  @author Kim Rutherford
 *  @version $Id: EntryEdit.java,v 1.127 2000/03/27 12:53:38 kmr Exp $
 *
 */

public class EntryEdit extends Frame
    implements SelectionOwner, GotoEventSource, FeatureDisplayOwner,
               EntryGroupChangeListener, EntryChangeListener {
  /**
   *  Create a new EntryEdit object and Frame.
   *  @param artemis_main The ArtemisMain object that created this EntryEdit.
   *  @param entry_group The EntryGroup object that this component is editing.
   */
  public EntryEdit (final ArtemisMain artemis_main,
                    final EntryGroup entry_group) {
    super ("Artemis Entry Edit");

    this.artemis_main = artemis_main;

    this.entry_group = entry_group;

//      if (Diana.isStandAlone ()) {
//        selection = new Selection (getToolkit().getSystemClipboard());
//      } else {
      selection = new Selection (null);
//      }

    getEntryGroup ().addFeatureChangeListener (selection);
    getEntryGroup ().addEntryChangeListener (selection);

    getEntryGroup ().addEntryGroupChangeListener (this);
    getEntryGroup ().addEntryChangeListener (this);

    if (getEntryGroup ().getDefaultEntry () != null) {
      final String name = getEntryGroup ().getDefaultEntry ().getName ();

      if (name != null) {
        setTitle ("Artemis Entry Edit: " + name);
      }
    }

    final int font_height;
    final int font_base_line;

    final Font default_font = getDefaultFont ();

    setFont (default_font);

    if (default_font != null) {
      FontMetrics fm = getFontMetrics (default_font);
      //      System.out.println ("fm:" + fm.toString ());
      font_height = fm.getHeight ();
      font_base_line = fm.getMaxAscent ();
    }
    else {
      font_height = -1;
      font_base_line = -1;
    }

//    System.out.println ("EntryEdit (): " + entry_group);

    addWindowListener (new WindowAdapter () {
      public void windowClosing (WindowEvent event) {
        entryEditFinished ();
      }
    });

    GridBagLayout gridbag = new GridBagLayout();
    setLayout (gridbag);

    menu_bar.setFont (default_font);

    GridBagConstraints c = new GridBagConstraints();

    c.gridwidth  = GridBagConstraints.REMAINDER;
    c.fill       = GridBagConstraints.HORIZONTAL;
    c.anchor     = GridBagConstraints.NORTH;
    c.gridheight = 1;
    c.weightx    = 1;
    c.weighty    = 0;

    // Create and add the stats_label.
    selection_info =
      new SelectionInfoDisplay (getEntryGroup (), getSelection ());
    gridbag.setConstraints (selection_info, c);
    add (selection_info);

    group_display = new EntryGroupDisplay (this);

    gridbag.setConstraints (group_display, c);
    add (group_display);

    feature_display =
      new FeatureDisplay (entry_group, getSelection (), this, this);

    base_plot_group = new BasePlotGroup (this);
    gridbag.setConstraints (base_plot_group, c);
    add (base_plot_group);
    feature_display.addDisplayAdjustmentListener (base_plot_group);

    gridbag.setConstraints (feature_display, c);
    add (feature_display);

    base_display =
      new FeatureDisplay (entry_group, getSelection (), this, this);
    base_display.setShowLabels (false);
    base_display.setScaleFactor (0);
    gridbag.setConstraints (base_display, c);
    add (base_display);

    final String list_option =
      Options.getOptions ().getProperty ("show_list");

    feature_list = new FeatureList (getEntryGroup (), getSelection (), this);
    feature_list.setFont (default_font);
    c.fill = GridBagConstraints.BOTH;
    c.weighty = 1;
    gridbag.setConstraints (feature_list, c);

    if (list_option == null || ! list_option.equals ("no")) {
      feature_list.setVisible (true);
    } else {
      feature_list.setVisible (false);
    }

    add (feature_list);

    makeMenus ();

    packme ();

    setSize (900,750);

    final Dimension screen = Toolkit.getDefaultToolkit().getScreenSize();
    setLocation (new Point ((screen.width - getSize ().width) / 2,
                            (screen.height - getSize ().height) / 2));
  }

  /**
   *  Redraw this component.  This method is public so that other classes can
   *  force an update if, for example, the options files is re-read.
   **/
  public void redisplay () {
    feature_display.redisplay ();
    base_display.redisplay ();
  }

  /**
   *  This method sends an GotoEvent to all the GotoEvent listeners that will
   *  make the given base visible.
   *  @param base_marker The base to make visible.
   **/
  public void gotoBase (final Marker base_marker) {
    final GotoEvent new_event = new GotoEvent (this, base_marker);

    sendGotoEvent (new_event);
  }

  /**
   *  This method sends an GotoEvent to all the GotoEvent listeners that will
   *  make the start of the first selected Feature or FeatureSegment visible.
   **/
  public void makeSelectionVisible () {
    final Marker first_base = getSelection ().getStartBaseOfSelection ();

    final GotoEvent new_event = new GotoEvent (this, first_base);

    sendGotoEvent (new_event);
  }

  /**
   *  Return an object that implements the GotoEventSource interface, and is
   *  the controlling object for Goto events associated with this object.
   **/
  public GotoEventSource getGotoEventSource () {
    return this;
  }

  /**
   *  Return an object that implements the SelectionOwner interface and holds
   *  the Selection for this object.
   **/
  public SelectionOwner getSelectionOwner () {
    return this;
  }

  /**
   *  Send the given event to all the GotoListeners.
   **/
  public void sendGotoEvent (final GotoEvent goto_event) {
    // don't send if there is nowhere to go to
    if (goto_event.getMarker () != null) {
      fireAction (goto_listener_list, goto_event);
    }
  }

  /**
   *  This method sends an GotoEvent to all the GotoEvent listeners that will
   *  make the given base visible.  This just calls makeBaseVisible ().
   *  @param destination_base The base to make visible.
   *  @return The reference of a MarkerRange object for the given base or null
   *    if the call fails.  It will fail only if the destination_base is less
   *    than 1 or greater than the length of the sequence.
   **/
  public MarkerRange gotoBase (final int destination_base) {
    try {
      final Strand forward_strand =
        getEntryGroup ().getBases ().getForwardStrand ();
      final Marker destination_marker =
        forward_strand.makeMarker (destination_base);
      gotoBase (destination_marker);
      return new MarkerRange (destination_marker.getStrand (),
                              destination_base, destination_base);
    } catch (OutOfRangeException e) {
      return null;
    }
  }

  /**
   *  This method sends a GotoEvent to all the GotoEvent listeners that will
   *  make the first base of the sequence visible.
   **/
  public void gotoFirstBase () {
    gotoBase (1);
  }

  /**
   *  This method sends a GotoEvent to all the GotoEvent listeners that will
   *  make the last base of the sequence visible.
   **/
  public void gotoLastBase () {
    gotoBase (getEntryGroup ().getSequenceLength ());
  }

  /**
   *  Adds the specified event listener to receive Goto events from this
   *  object.
   *  @param l the GotoEvent listener.
   **/
  public void addGotoListener (final GotoListener l) {
    goto_listener_list.addElement (l);
  }

  /**
   *  Removes the specified event listener so that it no longer receives Goto
   *  events from this object.
   *  @param l the GotoEvent listener.
   **/
  public void removeGotoListener (final GotoListener l) {
    goto_listener_list.removeElement (l);
  }

  /**
   *  Return the EntryGroup object that was passed to the constructor.  The
   *  EntryGroup can be changed by calling the methods EntryEdit.add () and
   *  EntryEdit.remove ().  An Object that needs to monitor changes to the
   *  EntryGroup should add itself as a listener to the EntryGroup.
   **/
  public EntryGroup getEntryGroup () {
    return entry_group;
  }

  /**
   *  Returns a Selection object containing the selected features/exons.
   **/
  public Selection getSelection () {
    return selection;
  }

  /**
   *  Implementation of the EntryGroupChangeListener interface.  We listen to
   *  EntryGroupChange events so that we can update the File menu when the
   *  EntryGroup changes.
   **/
  public void entryGroupChanged (final EntryGroupChangeEvent event) {
    switch (event.getType ()) {
    case EntryGroupChangeEvent.ENTRY_ADDED:
    case EntryGroupChangeEvent.ENTRY_DELETED:
      makeFileMenu ();
      break;
    }
  }

  /**
   *  Implementation of the EntryChangeListener interface.  We listen to
   *  EntryChange events so that we can update the File menu when an Entry
   *  changes.
   **/
  public void entryChanged (final EntryChangeEvent event) {
    if (event.getType () == EntryChangeEvent.NAME_CHANGED) {
      makeFileMenu ();
    }
  }

  /**
   *  Read a new entry into the entry group that this component is displaying.
   *  This method uses EntryFileDialog to open a requester.
   **/
  public void readEntry () {
    final EntryFileDialog dialog = new EntryFileDialog (this);

    // get the entry and add it to entry_group

    final uk.ac.sanger.pathogens.embl.Entry new_embl_entry;

    try {
      new_embl_entry = dialog.getEntry (Diana.getExtendedEntryInformation ());
    } catch (EntryInformationException e) {
      new MessageDialog (this, "read failed: " + e.getMessage ());
      return;
    }

    if (new_embl_entry == null) {
      // the read failed
      return;
    }

    final Bases embl_bases = new Bases (new_embl_entry.getSequence ());

    try {
      final Entry new_entry =
        new Entry (entry_group.getBases (), new_embl_entry);

      getEntryGroup ().add (new_entry);
    } catch (OutOfRangeException e) {
      new MessageDialog (this, "read failed: one of the features in " +
                         new_embl_entry.getName () +
                         " has an out of range location: " + e.getMessage ());
      return;
    }
  }

  /* private members: */

  /**
   *  This method arranges for the EntryEdit Frame to go away.  This EntryEdit
   *  object was created by the main program, so the main program must be the
   *  one to delete us.
   **/
  private void entryEditFinished () {
    getOwner ().entryEditFinished (this);
  }

  /**
   *  Send an event to those object listening for it.
   *  @param listeners A Vector of the objects that the event should be sent
   *    to.
   *  @param event The event to send
   **/
  private void fireAction (final Vector listeners, final EventObject event) {
    final Vector targets;
    // copied from a book - synchronising the whole method might cause a
    // deadlock
    synchronized (this) {
      targets = (Vector) listeners.clone ();
    }

    for (int i = 0 ; i < targets.size () ; ++i) {
      GotoListener target = (GotoListener) targets.elementAt (i);

      if (event instanceof GotoEvent) {
        final GotoListener goto_listener = (GotoListener) target;
        goto_listener.performGoto ((GotoEvent) event);
      } else {
        throw new Error ("EntryEdit.fireAction () - unknown event");
      }
    }
  }

  /**
   *  Write the default Entry in the EntryGroup to the file it came from.
   **/
  private void saveDefaultEntry () {
    if (getEntryGroup ().getDefaultEntry () == null) {
      new MessageDialog (EntryEdit.this, "There is no default entry");
    } else {
      saveEntry (entry_group.getDefaultEntry (), true, false, true,
                 DocumentEntryFactory.ANY_FORMAT);
    }
  }

  /**
   *  Save the given entry, prompting for a file name if necessary.
   *  @param include_diana_extensions If true then the any diana additions to
   *    the embl file format will be included in the output, otherwise they
   *    will be removed.  Also possible problems that would cause an entry to
   *    bounce from the EMBL submission process will be flagged if this is
   *    true.
   *  @param ask_for_name If true then always prompt for a new filename,
   *    otherwise prompt only when the entry name is not set.
   *  @param keep_new_name If ask_for_name is true a file will be written with
   *    the new name the user selects - if keep_new_name is true as well, then
   *    the entry will have it's name set to the new name, otherwise it will
   *    be used for this save and then discarded.
   *  @param destination_type Should be one of EMBL_FORMAT, GENBANK_FORMAT or
   *    ANY_FORMAT.  If ANY_FORMAT then the Entry will be saved in the
   *    same format it was created, otherwise it will be saved in the given
   *    format.
   **/
  void saveEntry (final Entry entry,
                  final boolean include_diana_extensions,
                  final boolean ask_for_name,
                  final boolean keep_new_name,
                  final int destination_type) {

    if (!include_diana_extensions) {
      if (displaySaveWarnings (entry)) {
        return;
      }
    }

    if (destination_type != DocumentEntryFactory.ANY_FORMAT &&
        entry.getHeaderText () != null) {
      final YesNoDialog yes_no_dialog =
        new YesNoDialog (this, "header section will be lost.  continue?");

      if (!yes_no_dialog.getResult ()) {
        return;
      }
    }

    try {
      if (ask_for_name || entry.getName () == null) {
        // prompt for a name
        final FileDialog dialog =
          new FileDialog (this,
                          "Select an EMBL file name...",
                          FileDialog.SAVE);

        dialog.setVisible (true);

        if (dialog.getFile () == null || dialog.getFile ().length () == 0) {
          new MessageDialog (this, "no file name given for save");
          return;
        } else {
          final String file_name = dialog.getDirectory () +
            java.io.File.separator + dialog.getFile ();

          if (new File (file_name).exists ()) {
            final YesNoDialog yes_no_dialog =
              new YesNoDialog (this, "this file exists: " + file_name +
                               " overwrite it?");

            if (yes_no_dialog.getResult ()) {
              // yes - continue
            } else {
              // no
              return;
            }
          }

          System.out.println ("trying to save to: " +  file_name);

          final MessageDialog message;

          if (entry.getFeatureCount () > 200) {
            message = new MessageDialog (this, "saving ...", false);
          } else {
            message = null;
          }

          try {
            if (include_diana_extensions) {
              entry.save (file_name, destination_type, false);
            } else {
              entry.saveStandardOnly (file_name, destination_type, true);
            }
          } catch (EntryInformationException e) {
            final YesNoDialog yes_no_dialog =
              new YesNoDialog (this, "destination format can't handle all " +
                               "keys/qualifiers continue?");
            if (yes_no_dialog.getResult ()) {
              try {
                if (include_diana_extensions) {
                  entry.save (file_name, destination_type, true);
                } else {
                  entry.saveStandardOnly (file_name, destination_type, true);
                }
              } catch (EntryInformationException e2) {
                throw new Error ("internal error - unexpected exception: "+ e);
              }
            } else {
              return;
            }
          }

          if (keep_new_name) {
            entry.setName (file_name);
          }

          if (message != null) {
            message.dispose ();
          }
        }
      } else {
        final MessageDialog message;

        if (entry.getFeatureCount () > 200) {
          message = new MessageDialog (this, "saving...", false);
        } else {
          message = null;
        }

        if (include_diana_extensions) {
          entry.save (destination_type);
        } else {
          entry.saveStandardOnly (destination_type);
        }

        if (message != null) {
          message.dispose ();
        }
      }
    } catch (ReadOnlyException e) {
      new MessageDialog (this, "this entry is read only");
      return;
    } catch (IOException e) {
      new MessageDialog (this, "error while saving: " + e);
      return;
    } catch (EntryInformationException e) {
      new MessageDialog (this, "error while saving: " + e);
      return;
    }
  }

  /**
   *  Save the changes to all the Entry objects in the entry_group back to
   *  where the they came from.
   **/
  public void saveAllEntries () {
    for (int entry_index = 0 ;
         entry_index < entry_group.size () ;
         ++entry_index) {
      saveEntry (entry_group.elementAt (entry_index), true, false, true,
                 DocumentEntryFactory.ANY_FORMAT);
    }
  }

  /**
   *  Check the given Entry for invalid EMBL features (such as CDS features
   *  without a stop codon) then display a FeatureListFrame list the problem
   *  features.
   *  @return true if and only if the save should be aborted.
   **/
  private boolean displaySaveWarnings (final Entry entry) {
    final FeatureVector invalid_starts = entry.checkFeatureStartCodons ();
    final FeatureVector invalid_stops = entry.checkFeatureStopCodons ();
    final FeatureVector invalid_keys = entry.checkForNonEMBLKeys ();
    final FeatureVector duplicate_features = entry.checkForEMBLDuplicates ();
    final FeatureVector overlapping_cds_features =
      entry.checkForOverlappingCDSs ();
    final FeatureVector missing_qualifier_features =
      entry.checkForMissingQualifiers ();

    // this predicate will filter out those features that aren't in the
    // entry we are trying to save
    final FeaturePredicate predicate =
      new FeaturePredicate () {
        public boolean testPredicate (final Feature feature) {
          if (feature.getEntry () == entry) {
            return true;
          } else {
            return false;
          }
        }
      };

    final FilteredEntryGroup filtered_entry_group =
      new FilteredEntryGroup (getEntryGroup (), predicate);

    if (invalid_starts.size () + invalid_stops.size () +
        invalid_keys.size () + duplicate_features.size () +
        overlapping_cds_features.size () > 0) {

      final YesNoDialog yes_no_dialog =
        new YesNoDialog (this,
                         "warning: some features may have problems. " +
                         "continue with save?");

      if (!yes_no_dialog.getResult ()) {
        getSelection ().clear ();

        if (invalid_starts.size () > 0) {
          getSelection ().add (invalid_starts);

          ViewMenu.showBadStartCodons (this,
                                       getSelection (),
                                       filtered_entry_group,
                                       this);
        }

        if (invalid_stops.size () > 0) {
          getSelection ().add (invalid_stops);

          ViewMenu.showBadStopCodons (this,
                                      getSelection (),
                                      filtered_entry_group,
                                      this);
        }

        if (invalid_keys.size () > 0) {
          getSelection ().add (invalid_keys);

          ViewMenu.showNonEMBLKeys (this,
                                    getSelection (),
                                    filtered_entry_group,
                                    this);
        }

        if (duplicate_features.size () > 0) {
          getSelection ().add (duplicate_features);

          ViewMenu.showDuplicatedFeatures (this,
                                           getSelection (),
                                           filtered_entry_group,
                                           this);
        }

        if (overlapping_cds_features.size () > 0) {
          getSelection ().add (overlapping_cds_features);

          ViewMenu.showOverlappingCDSs (this,
                                        getSelection (),
                                        filtered_entry_group,
                                        this);
        }

        if (missing_qualifier_features.size () > 0) {
          getSelection ().add (missing_qualifier_features);

          ViewMenu.showMissingQualifierFeatures (this,
                                                 getSelection (),
                                                 filtered_entry_group,
                                                 this);
        }

        return true;
      }
    }

    return false;
  }

  /**
   *  Read the features from a file choosen by the user into the given entry.
   *  This method uses EntryFileDialog to open a requester.
   **/
  public void readFeatures (final Entry destination_entry) {
    final EntryFileDialog dialog = new EntryFileDialog (EntryEdit.this);

    // get the entry and add it to entry_group
    final uk.ac.sanger.pathogens.embl.Entry new_embl_entry;

    try {
      new_embl_entry = dialog.getEntry (Diana.getExtendedEntryInformation ());
    } catch (EntryInformationException e) {
      new MessageDialog (this, "read failed: " + e.getMessage ());
      return;
    }

    if (new_embl_entry == null) {
      return;
    }

    final FeatureVector lost_features = new FeatureVector ();

    final Entry entry;

    try {
      entry = new Entry (getEntryGroup ().getBases (), new_embl_entry);

      final FeatureVector new_features = entry.getAllFeatures ();

      for (int i = 0 ; i < new_features.size () ; ++i) {
        final Feature this_feature = new_features.elementAt (i);

        try {
          this_feature.removeFromEntry ();
        } catch (ReadOnlyException e) {
          throw new Error ("internal error - unexpected exception: " + e);
        }

        try {
          destination_entry.add (this_feature, false);
        } catch (EntryInformationException e) {
          if (e instanceof InvalidKeyException) {
            // the feature can't be read
            lost_features.add (this_feature);
          } else {
            try {
              // add it anywhere (qualifiers will be lost)
              destination_entry.add (this_feature, true);
            } catch (EntryInformationException e2) {
              throw new Error ("internal error - unexpected exception: " + e2);
            } catch (ReadOnlyException e2) {
              new MessageDialog (EntryEdit.this, "the entry is read only");
            }
          }
        } catch (ReadOnlyException e) {
          new MessageDialog (EntryEdit.this, "the entry is read only");
        }
      }

      if (lost_features.size () > 0) {
        final StringBuffer buffer =
          new StringBuffer ("some features types could not be read: ");

        for (int i = 0 ; i < lost_features.size () ; ++i) {
          buffer.append (lost_features.elementAt (0).getIDString ());
          if (i == 2) {
            buffer.append (", ...");
            break;
          }
          if (i < lost_features.size () - 1) {
            buffer.append (", ");
          }
        }

        new MessageDialog (EntryEdit.this, buffer.toString ());
      }
    } catch (OutOfRangeException e) {
      new MessageDialog (EntryEdit.this,
                         "read failed: one of the features in " +
                         "the entry has an out of range location");
      return;
    }
  }

  /**
   *  This method will call pack () and then move the Frame to the centre of
   *  the screen.  (Implementation of the FeatureDisplayOwner interface).
   **/
  public void packme () {
    pack ();

    final Dimension screen = Toolkit.getDefaultToolkit().getScreenSize();

    final int x_position = (screen.width - getSize ().width) / 2;
    int y_position = (screen.height - getSize ().height) / 2;

    if (y_position < 10) {
      y_position = 10;
    }

    setLocation (new Point (x_position, y_position));
  }

  /**
   *  Return the ArtemisMain object that was passed to the constructor.
   **/
  public ArtemisMain getOwner () {
    return artemis_main;
  }

  /**
   *  Make and add the menus for this component.
   **/
  private void makeMenus () {
    final Font default_font = getDefaultFont ();

    setMenuBar (menu_bar);

    makeFileMenu ();

    menu_bar.add (file_menu);

    final CheckboxMenuItem show_entry_buttons_item =
      new CheckboxMenuItem ("Show Entry Buttons");
    show_entry_buttons_item.setState (true);
    show_entry_buttons_item.addItemListener (new ItemListener () {
      public void itemStateChanged(ItemEvent event) {
        group_display.setVisible (show_entry_buttons_item.getState ());
        packme ();
      }
    });
    display_menu.add (show_entry_buttons_item);

    final CheckboxMenuItem show_base_display_item =
      new CheckboxMenuItem ("Show Base View");
    show_base_display_item.setState (true);
    show_base_display_item.addItemListener (new ItemListener () {
      public void itemStateChanged(ItemEvent event) {
        base_display.setVisible (show_base_display_item.getState ());
        packme ();
      }
    });
    display_menu.add (show_base_display_item);

    final CheckboxMenuItem show_feature_list_item =
      new CheckboxMenuItem ("Show Feature List");
    show_feature_list_item.setState (feature_list.isVisible ());
    show_feature_list_item.addItemListener (new ItemListener () {
      public void itemStateChanged(ItemEvent event) {
        feature_list.setVisible (show_feature_list_item.getState ());
        packme ();
      }
    });
    display_menu.add (show_feature_list_item);


    display_menu.addSeparator ();

    if (Diana.isStandAlone ()) {
      final MenuItem usage_plot_item =
        new MenuItem ("Add Usage Plots ...");
      usage_plot_item.addActionListener (new ActionListener () {
        public void actionPerformed (ActionEvent event) {
          addUsagePlot ();
        }
      });
      display_menu.add (usage_plot_item);

      display_menu.addSeparator ();
    }


    final BaseAlgorithm [] algorithms = base_plot_group.getPlotAlgorithms ();

    for (int i = 0 ; i < algorithms.length ; ++i) {
      final BaseAlgorithm this_algorithm = algorithms[i];

      addDisplayMenuAlgorithm (this_algorithm);
    }

    entry_group_menu = new EntryGroupMenu (this, getEntryGroup ());
    menu_bar.add (entry_group_menu);

    select_menu =
      new SelectMenu (this, getSelection (), this, getEntryGroup ());
    menu_bar.add (select_menu);

    view_menu = new ViewMenu (this, getSelection (), this, getEntryGroup ());
    menu_bar.add (view_menu);

    goto_menu = new GotoMenu (this, getSelection (), this, getEntryGroup ());
    menu_bar.add (goto_menu);

    if (Diana.isStandAlone ()) {
      edit_menu = new EditMenu (this, getSelection (), this, getEntryGroup ());
      menu_bar.add (edit_menu);

      add_menu = new AddMenu (this, getSelection (), getEntryGroup ());
      menu_bar.add (add_menu);

      write_menu = new WriteMenu (this, getSelection (), getEntryGroup ());
      menu_bar.add (write_menu);

      if (System.getProperty ("artemis.environment") != null &&
          System.getProperty ("artemis.environment").equals ("UNIX")) {
        run_menu = new RunMenu (this, getSelection ());
        menu_bar.add (run_menu);
      }
    }

    menu_bar.add (display_menu);

    if (Options.getOptions ().getProperty ("codon_usage_file") != null) {
      final String codon_usage_file_name =
        Options.getOptions ().getProperty ("codon_usage_file");

      try {
        addUsagePlot (codon_usage_file_name, true, false);
        addUsagePlot (codon_usage_file_name, false, false);

        packme ();
      } catch (IOException e) {
        new MessageDialog (this, "error while reading usage data: " + e);
      }
    }
  }

  /**
   *  Make a new File menu replacing the current one (if any).
   **/
  private void makeFileMenu () {
    file_menu.removeAll ();

    if (Diana.isStandAlone ()) {
      // the applet can't save or read

      final Menu read_features_menu = new Menu ("Read Features Into");

      file_menu.add (read_features_menu);

      final MenuItem read_entry = new MenuItem ("Read An Entry ...");
      read_entry.addActionListener (new ActionListener () {
        public void actionPerformed (ActionEvent event) {
          readEntry ();
        }
      });

      file_menu.add (read_entry);


      final String db_ior_url =
        Options.getOptions ().getProperty ("db_ior_url");

      if (db_ior_url != null) {
        final MenuItem read_db_entry =
          new MenuItem ("Read A Database Entry ...");
        read_db_entry.addActionListener (new ActionListener () {
          public void actionPerformed (ActionEvent event) {
            getOwner ().makeCorbaDialog (getEntryGroup (),
                                         db_ior_url, false, false);
          }
        });

        file_menu.add (read_db_entry);
      }

      final String ensembl_ior_url =
        Options.getOptions ().getProperty ("ensembl_ior_url");

      if (ensembl_ior_url != null) {
        final MenuItem read_ensembl_entry =
          new MenuItem ("Read An EnsEMBL Entry ...");
        read_ensembl_entry.addActionListener (new ActionListener () {
          public void actionPerformed (ActionEvent event) {
            getOwner ().makeCorbaDialog (getEntryGroup (),
                                         db_ior_url, false, true);
          }
        });

        file_menu.add (read_ensembl_entry);
      }

      file_menu.addSeparator ();

      final MenuItem clone_entry_edit = new MenuItem ("Clone This Window");
      clone_entry_edit.addActionListener (new ActionListener () {
        public void actionPerformed (ActionEvent event) {
          getOwner ().makeEntryEdit (getEntryGroup ()).show ();
        }
      });

      file_menu.add (clone_entry_edit);

      file_menu.addSeparator ();

      final MenuItem save_default =
        new MenuItem ("Save Default Entry", new MenuShortcut (KeyEvent.VK_S));
      save_default.addActionListener (new ActionListener () {
        public void actionPerformed (ActionEvent event) {
          saveDefaultEntry ();
        }
      });

      file_menu.add (save_default);

      final Menu save_entry_menu = new Menu ("Save An Entry");

      final Menu save_as_menu = new Menu ("Save An Entry As");


      if (entry_group == null || entry_group.size () == 0) {
        read_features_menu.add (new MenuItem ("(No Entries Currently)"));
        save_entry_menu.add (new MenuItem ("(No Entries Currently)"));
        save_as_menu.add (new MenuItem ("(No Entries Currently)"));
      } else {
        final Menu save_as = new Menu ("New File");
        final Menu save_as_embl = new Menu ("EMBL Format");
        final Menu save_as_genbank = new Menu ("GENBANK Format");
        final Menu save_embl_only = new Menu ("EMBL Submission Format");

        for (int i = 0 ; i < getEntryGroup ().size () ; ++i) {
          final Entry this_entry = getEntryGroup ().elementAt (i);

          String entry_name = this_entry.getName ();

          if (entry_name == null) {
            entry_name = "no name";
          }

          final ActionListener read_into_listener =
            new ReadFeaturesActionListener (this, this_entry);

          final MenuItem read_into_item = new MenuItem (entry_name);

          read_into_item.addActionListener (read_into_listener);

          final ActionListener save_entry_listener =
            new SaveEntryActionListener (this, this_entry);

          final MenuItem save_entry_item = new MenuItem (entry_name);

          save_entry_item.addActionListener (save_entry_listener);

          final ActionListener save_as_listener =
            new SaveEntryAsActionListener (this, this_entry);

          final MenuItem save_as_item = new MenuItem (entry_name);

          save_as_item.addActionListener (save_as_listener);

          final ActionListener save_as_embl_listener =
            new SaveEntryAsEMBLActionListener (this, this_entry);

          final MenuItem save_as_embl_item = new MenuItem (entry_name);

          save_as_embl_item.addActionListener (save_as_embl_listener);

          final ActionListener save_as_genbank_listener =
            new SaveEntryAsGenbankActionListener (this, this_entry);

          final MenuItem save_as_genbank_item = new MenuItem (entry_name);

          save_as_genbank_item.addActionListener (save_as_genbank_listener);

          final ActionListener save_embl_only_listener =
            new SaveEntryAsSubmissionActionListener (this, this_entry);

          final MenuItem save_embl_only_item = new MenuItem (entry_name);

          save_embl_only_item.addActionListener (save_embl_only_listener);

          read_features_menu.add (read_into_item);

          save_entry_menu.add (save_entry_item);

          save_as.add (save_as_item);
          save_as_embl.add (save_as_embl_item);
          save_as_genbank.add (save_as_genbank_item);
          save_embl_only.add (save_embl_only_item);
        }

        save_as_menu.add (save_as);
        save_as_menu.add (save_as_embl);
        save_as_menu.add (save_as_genbank);
        save_as_menu.addSeparator ();
        save_as_menu.add (save_embl_only);
      }

      final MenuItem save_all = new MenuItem ("Save All Entries");
      save_all.addActionListener (new ActionListener () {
        public void actionPerformed (ActionEvent event) {
          saveAllEntries ();
        }
      });

      file_menu.add (save_entry_menu);

      file_menu.add (save_as_menu);

      file_menu.add (save_all);

      file_menu.addSeparator ();
    }

    final MenuItem close = new MenuItem ("Close");
    close.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        if (getEntryGroup ().hasUnsavedChanges ()) {
          final YesNoDialog yes_no_dialog =
            new YesNoDialog (EntryEdit.this,
                             "there are unsaved changes - really close?");

          if (!yes_no_dialog.getResult ()) {
            return;
          }
        }

        entryEditFinished ();
      }
    });

    file_menu.add (close);
  }

  /**
   *  Ask the user for a file name, then read the codon usage data from that
   *  file, then make and add forward and a reverse BasePlot component using
   *  the data.
   **/
  private void addUsagePlot () {
    final FileDialog dialog =
      new FileDialog (this,
                      "Select a codon usage data file name ...",
                      FileDialog.LOAD);

    dialog.setVisible (true);

    if (dialog.getFile () != null && dialog.getFile ().length () != 0) {
      final String codon_usage_file_name =
        dialog.getDirectory () + java.io.File.separator + dialog.getFile ();

      try {
        final BasePlot new_forward_plot =
          addUsagePlot (codon_usage_file_name, true, true);
        final BasePlot new_reverse_plot =
          addUsagePlot (codon_usage_file_name, false, true);

        final Algorithm forward_algorithm = new_forward_plot.getAlgorithm ();
        final Algorithm reverse_algorithm = new_reverse_plot.getAlgorithm ();

        base_plot_group.setVisibleByAlgorithm (forward_algorithm, true);
        base_plot_group.setVisibleByAlgorithm (reverse_algorithm, true);

        packme ();
      } catch (IOException e) {
        new MessageDialog (this, "error while reading usage data: " + e);
      }
    }
  }


  /**
   *  Read the codon usage data from the given file, then make and add a
   *  BasePlot component using the data.
   *  @param use_forward_strand The plot will be a forward plot if and only if
   *    this is true.
   *  @param is_visible The plot will start off visible if and only if this is
   *    true.
   *  @return The BasePlot that was added.
   **/
  private BasePlot addUsagePlot (final String codon_usage_file_name,
                                 final boolean use_forward_strand,
                                 final boolean is_visible)
      throws IOException {
    final CodonUsageAlgorithm new_algorithm;

    if (use_forward_strand) {
      final Strand forward_strand =
        getEntryGroup ().getBases ().getForwardStrand ();

      final CodonUsageWeight usage_weights =
        new CodonUsageWeight (codon_usage_file_name, forward_strand);

      new_algorithm =
        new CodonUsageAlgorithm (forward_strand, usage_weights);
    } else {
      final Strand backward_strand =
        getEntryGroup ().getBases ().getBackwardStrand ();

      final CodonUsageWeight usage_weights =
        new CodonUsageWeight (codon_usage_file_name, backward_strand);

      new_algorithm =
        new CodonUsageAlgorithm (backward_strand, usage_weights);
    }

    view_menu.setCodonUsageAlgorithm (new_algorithm);

    add_menu.setCodonUsageAlgorithm (new_algorithm);

    final BasePlot new_base_plot =
      base_plot_group.addAlgorithm (new_algorithm);

    base_plot_group.setVisibleByAlgorithm (new_algorithm, is_visible);

    addDisplayMenuAlgorithm (new_algorithm);

    // hack to force the BasePlot to initialise
    feature_display.fireAdjustmentEvent ();

    return new_base_plot;
  }

  /**
   *  Add a menu item for the given algorithm to the Display menu.
   **/
  private void addDisplayMenuAlgorithm (final BaseAlgorithm algorithm) {
    final CheckboxMenuItem new_item =
      new CheckboxMenuItem (algorithm.getAlgorithmName ());

    new_item.setState (base_plot_group.basePlotIsVisible (algorithm));

    new_item.addItemListener (new ItemListener () {
      public void itemStateChanged(ItemEvent event) {
        base_plot_group.setVisibleByAlgorithm (algorithm,
                                               new_item.getState ());
        packme ();
      }
    });

    display_menu.add (new_item);
  }

  /**
   *  Return the current default font (from Diana.options).
   **/
  private Font getDefaultFont () {
    return Options.getOptions ().getFont ();
  }

  /**
   *  A vector containing the Entry objects that this EntryEdit object knows
   *  about.  The first Entry object in the entry_group is special - see
   *  saveEntry () for details.
   **/
  private EntryGroup entry_group;

  private final Panel   top_panel    = new Panel ();
  private final MenuBar menu_bar     = new MenuBar ();
  private final Menu    file_menu    = new Menu ("File");
  private Menu          select_menu;
  private ViewMenu      view_menu;
  private Menu          goto_menu;
  private Menu          entry_group_menu;
  private Menu          edit_menu;
  private AddMenu       add_menu;
  private Menu          write_menu;
//  private Menu          options_menu;
  private Menu          run_menu;
  private Menu          graph_menu;
  private final Menu    display_menu = new Menu ("Display");

  private EntryGroupDisplay    group_display;
  private FeatureDisplay       feature_display;
  private FeatureDisplay       base_display;
  private BasePlotGroup        base_plot_group;
  private SelectionInfoDisplay selection_info;

  private FeatureList feature_list;

  /**
   *  This Object contains the current selection.
   **/
  private Selection selection = null;

  /**
   *  A vector of those objects listening for selection change events.
   **/
  final private Vector goto_listener_list = new Vector ();

  /**
   *  The ArtemisMain reference that was passed to the constructor.
   **/
  final ArtemisMain artemis_main;
}

/**
 *  This is an EntryActionListener that will call readFeatures () when
 *  actionPerformed () is called.
 **/
class ReadFeaturesActionListener extends EntryActionListener {
  ReadFeaturesActionListener (final EntryEdit entry_edit,
                              final Entry entry) {
    super (entry_edit, entry);
  }

  public void actionPerformed (final ActionEvent event) {
    getEntryEdit ().readFeatures (getEntry ());
  }
};

/**
 *  This is an EntryActionListener that will call saveEntry () when
 *  actionPerformed () is called.
 **/
class SaveEntryActionListener extends EntryActionListener {
  SaveEntryActionListener (final EntryEdit entry_edit,
                           final Entry entry) {
    super (entry_edit, entry);
  }

  public void actionPerformed (final ActionEvent event) {
    getEntryEdit ().saveEntry (getEntry (), true, false, true,
                               DocumentEntryFactory.ANY_FORMAT);
  }
};

/**
 *  This is an EntryActionListener that will call saveEntry () when
 *  actionPerformed () is called.
 **/
class SaveEntryAsActionListener extends EntryActionListener {
  SaveEntryAsActionListener (final EntryEdit entry_edit,
                             final Entry entry) {
    super (entry_edit, entry);
  }

  public void actionPerformed (final ActionEvent event) {
    getEntryEdit ().saveEntry (getEntry (), true, true, true,
                               DocumentEntryFactory.ANY_FORMAT);
  }
};

/**
 *  This is an EntryActionListener that will call saveEntry () when
 *  actionPerformed () is called.
 **/
class SaveEntryAsEMBLActionListener extends EntryActionListener {
  SaveEntryAsEMBLActionListener (final EntryEdit entry_edit,
                                 final Entry entry) {
    super (entry_edit, entry);
  }

  public void actionPerformed (final ActionEvent event) {
    getEntryEdit ().saveEntry (getEntry (), true, true, false,
                               DocumentEntryFactory.EMBL_FORMAT);
  }
};

/**
 *  This is an EntryActionListener that will call saveEntry () when
 *  actionPerformed () is called.
 **/
class SaveEntryAsGenbankActionListener extends EntryActionListener {
  SaveEntryAsGenbankActionListener (final EntryEdit entry_edit,
                                    final Entry entry) {
    super (entry_edit, entry);
  }

  public void actionPerformed (final ActionEvent event) {
    getEntryEdit ().saveEntry (getEntry (), true, true, false,
                               DocumentEntryFactory.GENBANK_FORMAT);
  }
};

/**
 *  This is an EntryActionListener that will call saveEntry () when
 *  actionPerformed () is called.
 **/
class SaveEntryAsSubmissionActionListener extends EntryActionListener {
  SaveEntryAsSubmissionActionListener (final EntryEdit entry_edit,
                                       final Entry entry) {
    super (entry_edit, entry);
  }

  public void actionPerformed (final ActionEvent event) {
    getEntryEdit ().saveEntry (getEntry (), false, true, false,
                               DocumentEntryFactory.ANY_FORMAT);
  }
};
