/* FeatureDisplay.java
 *
 * created: Fri Oct  9 1998
 *
 * This file is part of Artemis
 *
 * Copyright (C) 1998,1999,2000  Genome Research Limited
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /nfs/disk222/yeastpub/Repository/powmap/diana/components/FeatureDisplay.java,v 1.111 2000/03/20 10:26:49 kmr Exp $
 */

package diana.components;

import Diana;
import uk.ac.sanger.pathogens.*;
import uk.ac.sanger.pathogens.embl.Range;
import diana.*;
import diana.sequence.*;

import collections.Comparator;

import java.awt.event.*;
import java.awt.*;
import java.lang.Math;
import java.util.Vector;

/**
 *  This component is used for displaying an Entry.
 *
 *  @author Kim Rutherford
 *  @version $Id: FeatureDisplay.java,v 1.111 2000/03/20 10:26:49 kmr Exp $
 **/

public class FeatureDisplay extends Panel
  implements EntryGroupChangeListener,
             EntryChangeListener, FeatureChangeListener,
             SelectionChangeListener, GotoListener, SequenceChangeListener,
             KeyListener, DisplayComponent
{

  /**
   *  Create a new FeatureDisplay object with the horizontal scrollbar at the
   *  bottom of the component.
   *  @param entry_group The EntryGroup that this component will display.
   *  @param owning_component The EntryEdit object that contains the selection
   *    that this component uses.
   **/
  public FeatureDisplay (final EntryGroup entry_group,
                         final Selection selection,
                         final GotoEventSource goto_event_source,
                         final FeatureDisplayOwner owning_component) {
    this (entry_group, selection, goto_event_source, owning_component, false);
  }

  /**
   *  Create a new FeatureDisplay object.
   *  @param entry_group The EntryGroup that this component will display.
   *  @param owning_component The EntryEdit object that contains the selection
   *    that this component uses.
   *  @param scrollbar_at_top If true the horizontal scrollbar will be at the
   *    top of component.
   **/
  public FeatureDisplay (final EntryGroup entry_group,
                         final Selection selection,
                         final GotoEventSource goto_event_source,
                         final FeatureDisplayOwner owning_component,
                         final boolean scrollbar_at_top) {
    this.owning_component  = owning_component;
    this.entry_group       = entry_group;
    this.selection         = selection;
    this.goto_event_source = goto_event_source;

    final Font font = getDefaultFont ();

    setFont (font);
    setFontInfo ();

    setScaleValue ();

    setLayout(new BorderLayout ());

    // this creates map_panel, the canvas and a horizontal scrollbar
    createMapPanel (scrollbar_at_top);

    createScaleScrollbar ();

    updateVisibleFeatureVector ();

    getSelection ().addSelectionChangeListener (this);
    getGotoEventSource ().addGotoListener (this);

    getEntryGroup ().addEntryGroupChangeListener (this);
    getEntryGroup ().addEntryChangeListener (this);
    getEntryGroup ().addFeatureChangeListener (this);

    final int PRIORITY = Bases.LOW_PRIORITY;

    getEntryGroup ().getBases ().addSequenceChangeListener (this, PRIORITY);

    addKeyListener (this);
  }

  public final static int FORWARD = Bases.FORWARD;
  public final static int BACKWARD = Bases.BACKWARD;

  public final static int NO_FRAME         = FeatureSegment.NO_FRAME;
  public final static int FORWARD_STRAND   = FeatureSegment.FORWARD_STRAND;
  public final static int BACKWARD_STRAND  = FeatureSegment.BACKWARD_STRAND;
  public final static int FORWARD_FRAME_1  = FeatureSegment.FORWARD_FRAME_1;
  public final static int FORWARD_FRAME_2  = FeatureSegment.FORWARD_FRAME_2;
  public final static int FORWARD_FRAME_3  = FeatureSegment.FORWARD_FRAME_3;
  public final static int BACKWARD_FRAME_3 = FeatureSegment.BACKWARD_FRAME_3;
  public final static int BACKWARD_FRAME_2 = FeatureSegment.BACKWARD_FRAME_2;
  public final static int BACKWARD_FRAME_1 = FeatureSegment.BACKWARD_FRAME_1;
  public final static int SCALE_LINE       = FeatureSegment.SCALE_LINE;

  /**
   *  Returns the value of a flag that indicates whether this component can be
   *  traversed using Tab or Shift-Tab keyboard focus traversal - returns true
   *  for FeatureDisplay components
   **/
  public boolean isFocusTraversable () {
    return true;
  }

  /**
   *  Set value of the show label flag.
   *  @param show_label Show labels if and only if this argument is true.
   **/
  public void setShowLabels (boolean show_labels) {
    if (this.show_labels != show_labels) {
      this.show_labels = show_labels;
      fixDisplay ();
      getOwningComponent ().packme ();
    } else {
      // do nothing
    }
  }

  /**
   *  Get the value of the "show label" flag.
   **/
  public boolean getShowLabels () {
    return show_labels;
  }

  /**
   *  Set value of the show base colours flag.
   *  @param show_base_colours At scale_factor less than two show each base in
   *    a different colour if and only if this argument is true.
   **/
  public void setShowBaseColours (boolean show_base_colours) {
    if (this.show_base_colours != show_base_colours) {
      this.show_base_colours = show_base_colours;
      if (getScaleFactor () > 1) {
        setScaleFactor (1);
      }
      repaintCanvas ();
    } else {
      // do nothing
    }
  }

  /**
   *  Get the value of the "show base colours" flag.
   **/
  public boolean getShowBaseColours () {
    return show_base_colours;
  }

  /**
   *  Set value of the show stop codons flag.
   *  @param show_stop_codons Show stop codons if and only if this argument is
   *    true.
   **/
  public void setShowStopCodons (boolean show_stop_codons) {
    if (this.show_stop_codons != show_stop_codons) {
      this.show_stop_codons = show_stop_codons;
      repaintCanvas ();
    } else {
      // do nothing
    }
  }

  /**
   *  Return the value of the "show stop codons" flag.
   **/
  public boolean getShowStopCodons () {
    return show_stop_codons;
  }

  /**
   *  Set value of the show frame features flag.
   *  @param frame_features_flag All features (not just CDS features) will be
   *    drawn on the frame lines if and only if this argument is true.
   **/
  public void setFrameFeaturesFlag (boolean frame_features_flag) {
    if (this.frame_features_flag != frame_features_flag) {
      this.frame_features_flag = frame_features_flag;
      repaintCanvas ();
    } else {
      // do nothing
    }
  }

  /**
   *  Return the value of the "show frame features" flag.
   **/
  public boolean getFrameFeaturesFlag () {
    return frame_features_flag;
  }

  /**
   *  Set the value of the minimum score for this FeatureDisplay - features
   *  that have a /score lower than this value are never shown.
   **/
  public void setMinimumScore (final int minimum_score) {
    current_min_score = minimum_score;
    updateVisibleFeatureVector ();
    repaintCanvas ();
  }

  /**
   *  Return the value of the minimum score for this FeatureDisplay - see
   *  setMinimumScore ().
   **/
  public int getMinimumScore () {
    return current_min_score;
  }

  /**
   *  Set the value of the maximum score for this FeatureDisplay - features
   *  that have a /score higher than this value are never shown.
   **/
  public void setMaximumScore (final int maximum_score) {
    current_max_score = maximum_score;
    updateVisibleFeatureVector ();
    repaintCanvas ();
  }

  /**
   *  Return the value of the maximum score for this FeatureDisplay - see
   *  setMaximumScore ().
   **/
  public int getMaximumScore () {
    return current_max_score;
  }

  /**
   *  Redraw this component.  This method is public so that other classes can
   *  force an update.  for example this is called when the options files is
   *  re-read.
   **/
  public void redisplay () {
    repaintCanvas ();
  }

  /**
   *  Adds the specified event adjustment listener to receive adjustment
   *  change events from this object.
   *  @param l the event change listener.
   **/
  public void addDisplayAdjustmentListener (DisplayAdjustmentListener l) {
    adjustment_listener_list.addElement (l);
  }

  /**
   *  Removes the specified event listener so that it no longer receives
   *  adjustment change events from this object.
   *  @param l the event change listener.
   **/
  public void removeDisplayAdjustmentListener (DisplayAdjustmentListener l) {
    adjustment_listener_list.removeElement (l);
  }

  /**
   *  Set the scale factor and update the display if the scale factor has
   *  changed.  A factor of zero means the full translation will be visible.
   *  At higher scale factors only stop codons are visible, and a bigger
   *  number will mean more bases are visible.
   **/
  public void setScaleFactor (int scale_factor) {
    if (this.scale_factor != scale_factor) {
      // we will try to keep the base in the centre of the view to stay where
      // it is, so we save it's position in remember_position.

      int remember_position = getCentreForwardBase ();

      if (getFirstVisibleForwardBase () == 1) {
        // if the first base is visible then keep it visible
        remember_position = 1;
      }

      // but, we want to keep the selection visible after a scale change, so
      // that will override the centre position
      final Marker first_base_marker =
        getSelection ().getStartBaseOfSelection ();

      if (first_base_marker != null && baseVisible (first_base_marker)) {
        remember_position = first_base_marker.getRawPosition ();
      }

      final Marker last_base_marker =
        getSelection ().getStartBaseOfSelection ();

      if (last_base_marker != null && baseVisible (last_base_marker)) {
        remember_position = last_base_marker.getRawPosition ();
      }

      this.scale_factor = scale_factor;

      setScaleValue ();

      scale_changer.setValue (scale_factor);

      makeBaseVisible (remember_position, true);

      fixScrollbar ();
      fireAdjustmentEvent ();

      updateVisibleFeatureVector ();

      repaintCanvas ();
    }
  }

  /**
   *  Implementation of the KeyListener interface (does nothing).
   **/
  public void keyPressed (final KeyEvent e) {

  }

  /**
   *  Implementation of the KeyListener interface (does nothing).
   **/
  public void keyReleased (final KeyEvent e) {

  }

  /**
   *  Implementation of the KeyListener interface.
   **/
  public void keyTyped (final KeyEvent e) {
//    System.out.println ("key:" + e.getKeyChar () + "  " + e.getKeyCode ());
  }

   /**
   *  Implementation of the FeatureChangeListener interface.  We listen to
   *  FeatureChange events so that we can update the display if qualifiers
   *  change.
   **/
  public void featureChanged (final FeatureChangeEvent event) {
    final Feature event_feature = event.getFeature ();

    if (!getEntryGroup ().contains (event_feature)) {
      // the feature isn't in an active entry
      return;
    }

    // if the feature is visible now or is in the list of visible features
    // (ie. it was visible previously) then redisplay.
    if (featureVisible (event_feature) ||
        getVisibleFeatures ().contains (event_feature)) {

      // update the visible_features vector
      if (getVisibleFeatures ().contains (event_feature) &&
          !featureVisible (event_feature)) {
        getVisibleFeatures ().remove (event_feature);
        repaintCanvas ();
      } else {
        if (!getVisibleFeatures ().contains (event_feature) &&
            featureVisible (event_feature)) {
          getVisibleFeatures ().add (event_feature);
          repaintCanvas ();
        } else {
          // if the visibility of the feature hasn't changed then do a full
          // repaint only if the location has changed.  if the location hasn't
          // changed then just redraw the feature
          switch (event.getType ()) {
          case FeatureChangeEvent.QUALIFIER_CHANGED:
          case FeatureChangeEvent.KEY_CHANGED:
            drawFeature (canvas.getGraphics (), event_feature, true, true);
            break;
          default:
            repaintCanvas ();
            break;
          }
        }
      }
    }
  }

  /**
   *  Implementation of the EntryGroupChangeListener interface.  We listen to
   *  EntryGroupChange events so that we can update the display if entries
   *  are added or deleted.
   **/
  public void entryGroupChanged (final EntryGroupChangeEvent event) {
    switch (event.getType ()) {
    case EntryGroupChangeEvent.ENTRY_ADDED:
    case EntryGroupChangeEvent.ENTRY_ACTIVE:
    case EntryGroupChangeEvent.ENTRY_DELETED:
    case EntryGroupChangeEvent.ENTRY_INACTIVE:
      updateVisibleFeatureVector ();
      break;
    }

    repaintCanvas ();
  }

  /**
   *  Implementation of the EntryChangeListener interface.  We listen to
   *  EntryChange events so that we can update the display if features are
   *  added or deleted.
   **/
  public void entryChanged (final EntryChangeEvent event) {
//      System.out.println ("in FeatureDisplay entry change event type: " +
//                          event.getType () + " for feature " +
//                          event.getFeature ());

    switch (event.getType ()) {
    case EntryChangeEvent.FEATURE_DELETED:
      remove (event.getFeature ());
      break;
    case EntryChangeEvent.FEATURE_ADDED:
      add (event.getFeature ());
      break;
    }
  }

  /**
   *  Implementation of the SelectionChangeListener interface.  We listen to
   *  SelectionChange events so that we can update the list to reflect the
   *  current selection.
   **/
  public void selectionChanged (final SelectionChangeEvent event) {

    if (event.getSource () == this) {
      // don't bother with events we sent ourself
      return;
    }

    if (event.getType () == SelectionChangeEvent.SELECTION_CHANGED) {
      final FeatureVector all_features =
        getSelection ().getAllFeatures ();

      for (int i = 0 ; i < all_features.size () ; ++i) {
        raiseFeature (all_features.elementAt (i));
      }
    }

    repaintCanvas ();
  }

  /**
   *  Implementation of the SequenceChangeListener interface.
   **/
  public void sequenceChanged (final SequenceChangeEvent event) {
    visible_features = new FeatureVector ();
    makeBaseVisible (event.getPosition (), true);
    updateVisibleFeatureVector ();
    repaintCanvas ();
    fireAdjustmentEvent ();
  }

  /**
   *  Implementation of the GotoListener interface.  Invoked when the listener
   *  should goto to the given base.
   **/
  public void performGoto (final GotoEvent event) {
    makeBaseVisible (event.getMarker ());
  }

  /**
   *  Return a MarkerRange that exactly covers the visible bases
   **/
  public MarkerRange getVisibleRange () {
    final int first_base = getFirstVisibleForwardBase ();
    final int last_base  = getLastVisibleForwardBase ();

    final Strand strand = getEntryGroup ().getBases ().getForwardStrand ();

    try {
      return strand.makeMarkerRangeFromPositions (first_base, last_base);
    } catch (OutOfRangeException e) {
      throw new Error ("internal error - unexpected exception: " + e);
    }
  }

  /**
   *  Arrange for the given feature to be drawn last - ie. so that it will
   *  appear at the front of the other features.
   **/
  void raiseFeature (Feature feature) {
    if (getVisibleFeatures ().remove (feature)) {
      getVisibleFeatures ().addElementAtEnd (feature);
      repaintCanvas ();
    }
  }

  /**
   *  Arrange for the given feature to be drawn first - ie. so that it will
   *  appear at the back of the other features.
   **/
  void lowerFeature (Feature feature) {
    if (getVisibleFeatures ().remove (feature)) {
      getVisibleFeatures ().insertElementAt (feature, 0);
      repaintCanvas ();
    }
  }

  /**
   *  Redraw the display with the smallest features on top.  Selected features
   *  will always be on top.
   **/
  void smallestToFront () {
    visible_features = new FeatureVector ();
    updateVisibleFeatureVector ();
    repaintCanvas ();
  }

  /**
   *  Scroll the display so that the given base is in the middle of the screen
   *  @param base_position The base position to make visible.
   *  @param forward true means FORWARD - the base_position refers to a
   *    position on the forward strand, false means BACKWARD - the
   *    base_position refers to the reverse strand.
   **/
  private void makeBaseVisible (int base_position, boolean forward) {

    int forward_base_position = base_position;

    if (!forward) {
      final Bases bases = getEntryGroup ().getBases ();
      forward_base_position =
        bases.getComplementPosition (forward_base_position);
    }

    final int max_visible_bases = getMaxVisibleBases ();

    final int possible_base_position =
      forward_base_position - max_visible_bases / 2;

    int real_base_position;

    if (possible_base_position < 1) {
      real_base_position = 1;
    } else {
      real_base_position = possible_base_position;
    }

    if (real_base_position > getSequenceLength ()) {
      real_base_position = getSequenceLength ();
    }

    setFirstVisibleForwardBase (real_base_position);
  }

  /**
   *  Scroll the display so that the given base is in the middle of the screen
   *  @param base_marker The Marker of base to make visible.
   **/
  private void makeBaseVisible (Marker base_marker) {
    makeBaseVisible (base_marker.getPosition (),
                     base_marker.getStrand ().isForwardStrand ());
  }

  /**
   *  This method is called to add a single feature to the display by adding it
   *  to the vector of visible features (if it is visible that is).
   *  @param feature The object to add
   **/
  private void add (Feature feature) {
    if (getEntryGroup ().isActive (feature.getEntry ()) &&
        featureVisible (feature)) {

      if (visible_features.contains (feature)) {
//        throw new Error ("internal error - feature added a second time");
      } else {
        getVisibleFeatures ().addElementAtEnd (feature);
      }
    }

    repaintCanvas ();

    Thread.yield ();
  }


  /**
   *  The method removes a Feature from the this object.  The feature is
   *  removed from the vector of visible features and removed from the
   *  selection.
   *  @param feature The object to remove
   **/
  private void remove (Feature feature) {
//      System.out.println ("in FeatureDisplay.remove (): " + feature);

//      System.out.println ("in FeatureDisplay.remove (): " +
//                          getEntryGroup ().getDefaultEntry ().getFeatureCount ());

    if (visible_features != null) {
      visible_features.remove (feature);
    }

//    System.out.println (getVisibleFeatures ().size ());

    repaintCanvas ();

    Thread.yield ();
  }

  /**
   *  Returns an object containing the selected features.
   *  In future it might return entries or exons.
   **/
  public Selection getSelection  () {
    return selection;
  }

  /**
   *  Returns the vector containing those features that are currently visible.
   **/
  private FeatureVector getVisibleFeatures  () {
    return visible_features;
  }

  /**
   *  Returns a copy of the vector containing those features that are
   *  currently visible.
   **/
  public FeatureVector getCurrentVisibleFeatures  () {
    return (FeatureVector) visible_features.clone ();
  }

  /**
   *  This is called after scrolling to add all features that have become
   *  visible and remove those that have become invisible.  The method changes
   *  visible_features as little as possible so that features that were at the
   *  end (and hence drawn last/on top) stay there.  Selected features and
   *  segments will always be on top of unselected ones.
   **/
  private void updateVisibleFeatureVector () {
//      try {
//        throw new Error ("");
//      } catch (Error e) {
//        System.out.println ("updateVisibleFeatureVector ():");
//        e.printStackTrace ();
//      }

    final int first_visible_base = getFirstVisibleForwardBase ();
    final int last_visible_base = getLastVisibleForwardBase ();

    final FeatureVector real_visible_features =
      getSortedFeaturesInRange (newRange (first_visible_base,
                                          last_visible_base));

    final FeatureVector new_visible_features = new FeatureVector ();

    // add features that are in visible_features and
    // real_visible_features - ie features that are still visible
    for (int i = 0 ; i < visible_features.size () ; ++i) {
      final Feature new_feature = visible_features.elementAt (i);
      if (real_visible_features.contains (new_feature)) {
        new_visible_features.addElementAtEnd (new_feature);
      }
    }

    // add features that are in real_visible_features and not currently
    // in visible_features and are not selected (selected features will be
    // added last so that they stay on top).
    for (int i = 0 ; i < real_visible_features.size () ; ++i) {
      final Feature new_feature = real_visible_features.elementAt (i);

      if (!visible_features.contains (new_feature) &&
          !getSelection ().contains (new_feature)) {
        new_visible_features.addElementAtEnd (new_feature);
      }
    }

    final FeatureVector selection_features =
      getSelection ().getAllFeatures ();

    // now add features that are in real_visible_features, are not in
    // visible_features and are selected (selected features are added last so
    // that they stay on top).
    for (int i = 0 ; i < real_visible_features.size () ; ++i) {
      final Feature new_feature = real_visible_features.elementAt (i);
      if (!visible_features.contains (new_feature) &&
          selection_features.contains (new_feature)) {
        new_visible_features.addElementAtEnd (new_feature);
      }
    }

    visible_features = new_visible_features;
  }

  /**
   *  This is used by getSortedFeaturesInRange ().
   **/
  final private static Comparator feature_comparator = new Comparator () {
    /**
     *  Compare two Objects with respect to ordering.
     *  @return a negative number if feature1_object is less than
     *    feature2_object ; a positive number if feature1_object is greater
     *    than feature2_object; else 0
     **/
    public int compare(final Object feature1_object,
                       final Object feature2_object) {
      final Feature feature1 = (Feature) feature1_object;
      final Feature feature2 = (Feature) feature2_object;

      final int feature1_size = feature1.getBaseCount ();
      final int feature2_size = feature2.getBaseCount ();

      if (feature1_size > feature2_size) {
        return -1;
      } else {
        if (feature1_size < feature2_size) {
          return 1;
        } else {
          // use hash value as a last resort
          if (feature1.hashCode () < feature2.hashCode ()) {
            return -1;
          } else {
            if (feature1.hashCode () == feature2.hashCode ()) {
              return 0;
            } else {
              return 1;
            }
          }
        }
      }
    }
  };

  /**
   *  Return a vector containing the references of the Feature objects in the
   *  EntryGroup that are within the given range.
   *  @param range Return features that overlap this range - ie the start of
   *    the feature is less than or equal to the end of the range and the end
   *    of the feature is greater than or equal to the start of the range.
   *  @return The features the are within the given range.  The returned
   *    object is a copy - changes will not effect the EntryGroup object
   *    itself.  The Features are sorted so that the features with the least
   *    bases come last, which is the reverse of Entry.getFeaturesInRange ().
   **/
  private FeatureVector getSortedFeaturesInRange (final Range range) {
    final FeatureVector features_from_entry;

    try {
      features_from_entry = getEntryGroup ().getFeaturesInRange (range);

      final int min_score = getMinimumScore ();
      final int max_score = getMaximumScore ();

      if (min_score > 0 || max_score < 100) {
        // filter out low and high scoring features
        for (int i = features_from_entry.size () - 1 ; i >= 0 ; --i) {
          final int this_feature_score =
            features_from_entry.elementAt (i).getScore ();

          // features with no /score are always shown
          if (this_feature_score != -1 &&
              (this_feature_score < getMinimumScore () ||
               this_feature_score > getMaximumScore ())) {
            features_from_entry.removeElementAt (i);
          }
        }
      }
    } catch (OutOfRangeException e) {
      throw new Error ("internal error - unexpected exception: " + e);
    }

    return features_from_entry.sort (feature_comparator);
  }

  /**
   *  The main paint function for the canvas.  An off screen image used for
   *  double buffering when drawing the canvas.
   *  @param g The Graphics object of the canvas.
   **/
  private void paintCanvas (Graphics g) {
    synchronized (getEntryGroup ()) {
      final int canvas_width = canvas.getSize ().width;
      final int canvas_height = canvas.getSize ().height;

      if (canvas_height <= 0 || canvas_width <= 0) {
        // there is no point painting a zero width canvas
        return;
      }

      //    System.out.println ("in paintCanvas ()\n");

      if (offscreen == null) {
        //       System.out.println ("x: " + canvas.getSize ().width + " " +
        //                           "y: " + canvas.getSize ().height);

        offscreen = createImage (canvas_width,
                                 canvas_height);

        fixScrollbar ();
        updateVisibleFeatureVector ();
      }

      Graphics og = offscreen.getGraphics ();
      og.setClip (0, 0, canvas_width, canvas_height);

      fillBackground (og);

//        System.out.println ("size: " + visible_features.size ());

      // we draw the feature backgrounds first then the visible indication
      // that there is a MarkerRange selected, then the feature outlines.
      for (int i = 0 ; i < getVisibleFeatures ().size () ; ++i) {
        //         if (i%100 == 0) {
        //           System.out.println ("visible feature: " + i);
        //         }
        drawFeature (og, getVisibleFeatures ().elementAt (i), true, false);
      }

      drawBaseSelection (og);

      drawScale (og);

      drawCodons (og);

      Thread.yield ();

      drawBases (og);

      // now draw the feature outlines
      for (int i = 0 ; i < getVisibleFeatures ().size () ; ++i) {
        drawFeature (og, getVisibleFeatures ().elementAt (i), false, true);
      }

      g.drawImage (offscreen, 0, 0, null);
      og.dispose ();
    }
  }

  /**
   *  Draw the background colour of the frames.
   **/
  private void fillBackground (Graphics g) {
    g.setColor (Color.white);

    g.fillRect (0, 0, canvas.getSize ().width, canvas.getSize ().height);

    fillLane (g, FORWARD_FRAME_1,  lighter_grey);
    fillLane (g, FORWARD_FRAME_2,  lighter_grey);
    fillLane (g, FORWARD_FRAME_3,  lighter_grey);

    fillLane (g, FORWARD_STRAND,   not_so_light_grey);
    fillLane (g, BACKWARD_STRAND,  not_so_light_grey);

    fillLane (g, BACKWARD_FRAME_1, lighter_grey);
    fillLane (g, BACKWARD_FRAME_2, lighter_grey);
    fillLane (g, BACKWARD_FRAME_3, lighter_grey);
  }

  /**
   *  Fill one lane/frame line with the given colour.
   **/
  private void fillLane (Graphics g, int line_id, Color colour) {
    final int fill_line_number = getFrameDisplayLine (line_id);
    final int fill_line_top = fill_line_number * getLineHeight () + 1;

    g.setColor (colour);

    final int last_visible_base_coord =
      getHighXPositionOfBase (getLastVisibleForwardBase ());

    g.fillRect (0, fill_line_top,
                last_visible_base_coord, getFeatureHeight ());
  }

  // the minimum distance in pixels between the labels
  private final static int MINIMUM_LABEL_SPACING = 80;

  /**
   *  Draw the line showing the base numbers in the middle of the canvas.  The
   *  smallest numbers will be to the left.
   **/
  private void drawScale (Graphics g) {
    g.setColor (Color.black);

    final int canvas_width = canvas.getSize ().width;

    final int scale_line = getFrameDisplayLine (SCALE_LINE);
    final int scale_number_y_pos = scale_line * getLineHeight ();

    final float bases_per_pixel = (float) getMaxVisibleBases () / canvas_width;

    final int base_label_spacing;

    if (getScaleFactor () == 0) {
      // set the spacing so that the labels are at multiples of ten at the
      // lowest scale factor
      base_label_spacing =
        (int) Math.ceil (MINIMUM_LABEL_SPACING * bases_per_pixel / 10) * 10;
    } else {
      // set the spacing so that the labels are at multiples of 100 otherwise
      base_label_spacing =
        (int) Math.ceil (MINIMUM_LABEL_SPACING * bases_per_pixel / 100) * 100;
    }

    final int label_spacing = (int)(base_label_spacing / bases_per_pixel);

    final int possible_index_of_first_label =
      getFirstVisibleForwardBase () / base_label_spacing;

    final int index_of_first_label;

    if (possible_index_of_first_label == 0) {
      index_of_first_label = 1;
    } else {
      index_of_first_label = possible_index_of_first_label;
    }

    final int index_of_last_label =
      getLastVisibleForwardBase () / base_label_spacing;

//     System.out.println ("labels: " +
//                         (index_of_last_label - index_of_first_label) + "  " +
//                         index_of_first_label);


    for (int i = index_of_first_label ;
         i <= index_of_last_label ;
         ++i) {
      final String label_string =
        String.valueOf ((int)(i * base_label_spacing));

      final int scale_number_x_pos =
        getLowXPositionOfBase (i * base_label_spacing);

      g.drawString (label_string,
                    scale_number_x_pos + 2,
                    scale_number_y_pos + font_base_line);

      g.drawLine (scale_number_x_pos, scale_number_y_pos,
                  scale_number_x_pos, scale_number_y_pos + getLineHeight ());

      g.drawLine (scale_number_x_pos, scale_number_y_pos,
                  scale_number_x_pos + getFontWidth (), scale_number_y_pos);

    }
  }

  /**
   *  Draw the bases of the forward and backward strands into a Graphics
   *  object.
   **/
  private void drawBases (Graphics g) {
    if (getScaleFactor () > 1 ||
        getScaleFactor () == 1 && !show_base_colours) {
      return;
    }

    final Strand forward_strand =
      getEntryGroup ().getBases ().getForwardStrand ();
    final Strand backward_strand =
      getEntryGroup ().getBases ().getBackwardStrand ();

    final Range forward_range =
      newRange (getFirstVisibleForwardBase (),
                getLastVisibleForwardBase ());

    String forward_visible_bases =
      forward_strand.getSubSequence (forward_range).toUpperCase ();

    final int forward_frame_line =
      getFrameDisplayLine (FORWARD_STRAND);

    final int forward_sequence_length = forward_visible_bases.length ();

    for (int base_index = 0 ;
         base_index < forward_sequence_length ;
         ++base_index) {

      final char char_to_draw =
        forward_visible_bases.charAt (base_index);

      if (getScaleFactor () == 0) {
        drawOneBaseLetter (g,
                           char_to_draw,
                           base_index * getFontWidth (),
                           forward_frame_line * getLineHeight ());
      } else {
        drawOnePixelBase (g,
                          char_to_draw,
                          base_index,
                          forward_frame_line * getLineHeight ());
      }
    }

    final Range backward_range = newRange (getFirstVisibleBackwardBase (),
                                           getLastVisibleBackwardBase ());

    String backward_visible_bases =
      backward_strand.getSubSequence (backward_range).toUpperCase ();

    final int backward_frame_line =
      getFrameDisplayLine (BACKWARD_STRAND);

    final int backward_sequence_length = backward_visible_bases.length ();

    for (int base_index = 0 ;
         base_index < backward_sequence_length ;
         ++base_index) {

      final char char_to_draw =
        backward_visible_bases.charAt (backward_sequence_length -
                                       base_index - 1);

      if (getScaleFactor () == 0) {
        drawOneBaseLetter (g,
                           char_to_draw,
                           base_index * getFontWidth (),
                           backward_frame_line * getLineHeight ());
      } else {
        drawOnePixelBase (g,
                          char_to_draw,
                          base_index,
                          backward_frame_line * getLineHeight ());
      }
    }
  }


  /**
   *  Draw the codons of the sequence into a graphics object.  If the scale is
   *  0 then the codon letters will be drawn, otherwise just the stop codons
   *  will marked.
   *  @param g The object to draw into.
   **/
  private void drawCodons (Graphics g) {
    g.setColor (Color.black);

    if (getScaleFactor () == 0) {
      drawForwardCodons (g);
      drawBackwardCodons (g);
    } else {
      if (!show_stop_codons) {
        return;
      }

      drawForwardStopCodons (g);
      drawBackwardStopCodons (g);
    }
  }

  /**
   *  Mark the stop codons into the three forward frame lines.
   *  @param g The object to draw into.
   **/
  private void drawForwardStopCodons (Graphics g) {
    final Strand forward_strand =
      getEntryGroup ().getBases ().getForwardStrand ();

    final int first_visible_base = getFirstVisibleForwardBase ();

    final int frame_shift = (first_visible_base - 1) % 3;

    // base to start translation at - we start slightly off the left of the
    // screen
    final int start_base = first_visible_base - frame_shift;

    // base to end translation at
    // we + 3 to the upper bound because partial codons do not get translated
    // by getTranslation ()
    final int end_base = getLastVisibleForwardBase () + 3;

    final int [][] forward_stop_codons = {
      forward_strand.getStopCodons (newRange (start_base, end_base)),
      forward_strand.getStopCodons (newRange (start_base + 1, end_base)),
      forward_strand.getStopCodons (newRange (start_base + 2, end_base))
    };

    for (int i = 0 ; i < 3 ; ++i) {
      final int frame_line =
        getFrameDisplayLine (FORWARD_FRAME_1 + i);

      final int frame_x_start = i;

      final int [] this_frame_stop_codons = forward_stop_codons [i];

//      System.out.println ("drawForwardStopCodons (): " +
//                          frame_x_start);

      drawStopCodonLine (g,
                         frame_line,
                         this_frame_stop_codons,
                         FORWARD);
    }
  }

  /**
   *  Mark the stop codons into the three backframe frame lines.
   *  @param g The object to draw into.
   **/
  private void drawBackwardStopCodons (Graphics g) {
    final Strand backward_strand =
      getEntryGroup ().getBases ().getBackwardStrand ();

    final int first_visible_base = getFirstVisibleBackwardBase ();

    final int frame_shift = (first_visible_base - 1) % 3;

    // base to start translation at
    final int start_base = first_visible_base - frame_shift;

    // base to end translation at
    // we + 3 to the upper bound because partial codons do not get translated
    // by getTranslation ()
    final int end_base = getLastVisibleBackwardBase () + 3;

    final int [][] backward_stop_codons = {
      backward_strand.getStopCodons (newRange (start_base, end_base)),
      backward_strand.getStopCodons (newRange (start_base + 1, end_base)),
      backward_strand.getStopCodons (newRange (start_base + 2, end_base))
    };

    for (int i = 0 ; i < 3 ; ++i) {
      final int frame_line =
        getFrameDisplayLine (BACKWARD_FRAME_1 - i);

      final int frame_x_start = i;

      final int [] this_frame_stop_codons = backward_stop_codons [i];

      drawStopCodonLine (g,
                         frame_line,
                         this_frame_stop_codons,
                         BACKWARD);
    }
  }

  /**
   *  Draw the codon letters into the three forward frame lines.
   *  @param g The object to draw into.
   **/
  private void drawForwardCodons (Graphics g) {
    final Strand forward_strand =
      getEntryGroup ().getBases ().getForwardStrand ();

    final int first_visible_base = getFirstVisibleForwardBase ();

    for (int i = 0 ; i < 3 ; ++i) {
      final int frame_shift = 1 - (first_visible_base + 3 - i) % 3;

      // base to start translation at - we start slightly off the left of the
      // screen so that the first partial codon is translated as '.'
      final int start_base = first_visible_base + frame_shift;

      // base to end translation at - we end slightly off the right of the
      // screen
      final int end_base = getLastVisibleForwardBase () + 1;


      final int frame_line = getFrameDisplayLine (FORWARD_FRAME_1 + i);

      int frame_x_start = frame_shift;

      final AminoAcidSequence this_frame_translation =
        forward_strand.getTranslation (newRange (start_base, end_base));

      final String this_frame_translation_string =
        this_frame_translation.toString ();

      drawCodonLine (g,
                     frame_x_start,
                     frame_line,
                     this_frame_translation_string,
                     FORWARD);
    }
  }

  /**
   *  Draw the codon letters into the three backframe frame lines.
   *  @param g The object to draw into.
   **/
  private void drawBackwardCodons (Graphics g) {
    final Strand backward_strand =
      getEntryGroup ().getBases ().getBackwardStrand ();

    final int first_visible_base = getFirstVisibleBackwardBase ();

    for (int i = 0 ; i < 3 ; ++i) {
      final int frame_shift = 1 - (first_visible_base + 3 - i) % 3;

      // base to start translation at - we start slightly off the right of the
      // screen so that the first partial codon is translated as '.'
      final int start_base = first_visible_base + frame_shift;

      // base to end translation at - we end slightly off the left of the
      // screen
      final int end_base = getLastVisibleBackwardBase () + 1;

      final int frame_line = getFrameDisplayLine (BACKWARD_FRAME_1 - i);

      int frame_x_start = frame_shift;

      final AminoAcidSequence this_frame_translation =
        backward_strand.getTranslation (newRange (start_base, end_base));

      final String this_frame_translation_string =
        this_frame_translation.toString ();

      drawCodonLine (g,
                     frame_x_start,
                     frame_line,
                     this_frame_translation_string,
                     BACKWARD);
    }
  }


  /**
   *  Draw one line of codons.
   *  @param g The object to draw into.
   *  @param frame_start When drawing in the FORWARD direction this is the
   *    offset from the left of the screen at which to start drawing the line.
   *    For BACKWARD it is the offset from the last visible base.
   *  @param line_number The frame line to draw into.  (see
   *    getFrameDisplayLine () for more details about these line numbers).
   *  @param codons A String containing the codon letters to draw.
   *  @param direction The direction to draw the letters in (see the
   *    frame_start parameter).
   **/
  private void drawCodonLine (Graphics g,
                              int frame_start,
                              int line_number,
                              String codons,
                              int direction) {

//    System.out.println ("drawCodonLine: " + frame_start + " " + codons);

    final String upper_case_codons = codons.toUpperCase ();

    final int draw_y_position = line_number * getLineHeight ();

    // draw each AA letter
    for (int i = 0 ; i < upper_case_codons.length () ; ++i) {

      final char aa_char = upper_case_codons.charAt (i);

      int draw_x_position;

      draw_x_position = (int)((frame_start + 3 * i + 1) * getScaleValue ());

      if (direction == BACKWARD) {
        // draw in the opposite direction
        draw_x_position =
          getLowXPositionOfBase (getLastVisibleForwardBase ()) -
          draw_x_position;
      }

      drawOneLetter (g, aa_char, draw_x_position, draw_y_position);
    }
  }

  /**
   *  Draw one line of stop codons.
   *  @param g The object to draw into.
   *  @param line_number The frame line to draw into.  (see
   *    getFrameDisplayLine () for more details about these line numbers).
   *  @param stop_codons A vector containing the base positions of the stop
   *    codons.
   *  @param direction The direction to draw the letters in (see the
   *    frame_start parameter).
   **/
  private void drawStopCodonLine (Graphics g,
                                  int line_number,
                                  int [] stop_codons,
                                  int direction) {

//    System.out.println ("drawStopCodonLine (): " + stop_codons.size ());

    final int first_visible_forward_base = getFirstVisibleForwardBase ();
    final int first_visible_backward_base = getFirstVisibleBackwardBase ();
    final int last_visible_forward_base = getLastVisibleForwardBase ();
    final double scale_value = getScaleValue ();

    final int draw_y_position = line_number * getLineHeight ();

    // draw each stop codon
    for (int i = 0 ; i < stop_codons.length ; ++i) {

      final int codon_base_start_temp = stop_codons[i];

      if (codon_base_start_temp == 0) {
        // zero is the end of data marker
        break;
      }

      final boolean illegal_codon_flag = false;;

      final int codon_base_start = codon_base_start_temp;

// disabled for now:

//        if (codon_base_start_temp < 0) {
//          // a negative number in the stop_codons array indicates an
//          // illegal codon (one containing a non-base character)

//          codon_base_start = - codon_base_start_temp;
//          illegal_codon_flag = true;
//        } else {
//          illegal_codon_flag = false;
//        }

      // the number of bases difference between the codon_base_start and the
      // base at the left of the screen (right of screen for the backward
      // direction)
      final int codon_base_offset;

      if (direction == FORWARD) {
        codon_base_offset = codon_base_start - first_visible_forward_base;
      } else {
        codon_base_offset = codon_base_start - first_visible_backward_base;
      }

      int draw_x_position;

      draw_x_position = (int)(Math.floor (codon_base_offset * scale_value));

      if (direction == BACKWARD) {
        // draw in the opposite direction
        draw_x_position =
          getLowXPositionOfBase (last_visible_forward_base) - draw_x_position;
      } else {
//           System.out.println ("drawStopCodonLine (): " + frame_start + " " +
//                               codon_base_offset + " " +
//                               draw_x_position + " " +
//                               stop_codons[i] + " " + i);
      }

      drawOneStopCodon (g, draw_x_position, draw_y_position,
                        illegal_codon_flag, direction);
    }
  }

  /**
   *  Used by getColourOfBase () - this is the colour used for A.
   **/
  private static Color dark_green = new Color (0, 150, 0);

  /**
   *  Return the colour in which the given base should be drawn: blue for C,
   *  red for T, green for A and black for G.
   **/
  private Color getColourOfBase (final char base_char) {
    switch (base_char) {
    case 'c': case 'C':
      return Color.blue;
    case 't': case 'T':
      return Color.red;
    case 'a': case 'A':
      return dark_green;
    case 'g': case 'G':
      return Color.black;
    default:
      return Color.white;
    }
  }

  /**
   *  Used by drawOneLetter () - declared here so that we we don't need to
   *  allocated a whole array each time drawOneLetter () is called.
   **/
  private final char [] draw_one_char_temp_array = new char [1];


  /**
   *  Draw a letter at a specified position in a Graphics object.
   *  @param g The object to draw into.
   *  @param letter The character to draw.
   *  @param x_pos The x position at which to draw the letter.
   *  @param y_pos The y position at which to draw the letter.
   **/
  private void drawOneLetter (final Graphics g, final char letter,
                              final int x_pos, final int y_pos) {
//    System.out.println ("char: " + aa_char);

    draw_one_char_temp_array[0] = letter;

    g.drawChars (draw_one_char_temp_array, 0, 1,
                 x_pos,
                 y_pos + font_base_line);
  }

  /**
   *  Draw a letter at a specified position in a Graphics object.
   *  @param g The object to draw into.
   *  @param base_char The character of the base to draw (A, T, G or C).
   *  @param x_pos The x position at which to draw the letter.
   *  @param y_pos The y position at which to draw the letter.
   **/
  private void drawOneBaseLetter (final Graphics g, final char base_char,
                                  final int x_pos, final int y_pos) {
//    System.out.println ("char: " + base_char);

    if (show_base_colours) {
      g.setColor (getColourOfBase (base_char));
    }

    drawOneLetter (g, base_char, x_pos, y_pos);
  }


  /**
   *  Draw a stop codon at the given position.  The stop codon is represented
   *  by a one pixel wide line for scale factor greater than 1 and 3 pixels
   *  wide if the scale factor is one.
   *  @param g The object to draw into.
   *  @param x_pos The x position at which to draw the stop codon.
   *  @param y_pos The y position at which to draw the stop codon.  The line
   *    is drawn down from the given x,y position.
   *  @param illegal_codon_flag If true then draw a marker indicating that
   *    this codon is an illegal codon rather than a stop codon.
   *  @param direction If this is FORWARD the line is draw from the y position
   *    to the right, otherwise it is drawn to the left.
   **/
  private void drawOneStopCodon (final Graphics g,
                                 final int x_pos, final int y_pos,
                                 final boolean illegal_codon_flag,
                                 final int direction) {
    final int line_height;

    if (illegal_codon_flag) {
      g.setColor (Color.red);
      line_height = getLineHeight () * 2;
    } else {
      g.setColor (Color.black);
      line_height = getLineHeight ();
    }

    if (getScaleFactor () == 1) {
      // draw a line three pixels/bases wide
      if (direction == FORWARD) {
        g.drawLine (x_pos + 1, y_pos,
                    x_pos + 1, y_pos + line_height - 1);
        g.drawLine (x_pos + 2, y_pos,
                    x_pos + 2, y_pos + line_height - 1);
      } else {
        g.drawLine (x_pos - 1, y_pos,
                    x_pos - 1, y_pos + line_height - 1);
        g.drawLine (x_pos - 2, y_pos,
                    x_pos - 2, y_pos + line_height - 1);
      }
    }

    g.drawLine (x_pos, y_pos, x_pos, y_pos + line_height - 1);
  }

  /**
   *  Draw a one pixel wide line representing a single base on the forward or
   *  backward strand.  The base is coloured according to which base is passed
   *  to the method: blue for C, red for T, green for A and black for G.
   *  @param g The object to draw into.
   *  @param aa_char The character to draw.
   *  @param x_pos The x position at which to draw the line representing the
   *    base.
   *  @param y_pos The y position at which to draw the top of the line
   *    representing the base.
   **/
  private void drawOnePixelBase (Graphics g, char base_char,
                                 int x_pos, int y_pos) {
    g.setColor (getColourOfBase (base_char));

//      System.out.println ("drawOnePixelBase (): " + base_char + " " + x_pos +
//                          " " + y_pos);

    g.drawLine (x_pos, y_pos, x_pos, y_pos + getLineHeight () - 1);
  }

  /**
   *  Return the line on the canvas where this feature segment should be drawn.
   *  @param segment The segment in question.
   *  @return The line to draw into.
   **/
  private int getSegmentDisplayLine (FeatureSegment segment) {
    final int feature_frame_id = segment.getFrameID (getFrameFeaturesFlag ());

    return getFrameDisplayLine (feature_frame_id);
  }

  /**
   *  Return the line on the canvas where this frame ID should be drawn.  Call
   *  getLineCount () to find out the total number of lines.  Each line is
   *  getLineHeight () pixels high.
   *  @param frame_id The frame ID.
   *  @return The line to draw into.
   **/
  private int getFrameDisplayLine (int frame_id) {
    final int line_number;

    switch (frame_id) {
    case FORWARD_FRAME_1:
      if (show_labels) {
        line_number = 0;
      } else {
        line_number = 0;
      }
      break;
    case FORWARD_FRAME_2:
      if (show_labels) {
        line_number = 2;
      } else {
        line_number = 1;
      }
      break;
    case FORWARD_FRAME_3:
      if (show_labels) {
        line_number = 4;
      } else {
        line_number = 2;
      }
      break;
    case FORWARD_STRAND:
      if (show_labels) {
        line_number = 6;
      } else {
        line_number = 3;
      }
      break;
    case BACKWARD_STRAND:
      if (show_labels) {
        line_number = 9;
      } else {
        line_number = 5;
      }
      break;
    case BACKWARD_FRAME_1:
      if (show_labels) {
        line_number = 11;
      } else {
        line_number = 6;
      }
      break;
    case BACKWARD_FRAME_2:
      if (show_labels) {
        line_number = 13;
      } else {
        line_number = 7;
      }
      break;
    case BACKWARD_FRAME_3:
      if (show_labels) {
        line_number = 15;
      } else {
        line_number = 8;
      }
      break;
    default:
      if (show_labels) {
        line_number = 8;
      } else {
        line_number = 4;
      }
      break;
    }

    return line_number;
  }

  /**
   *  Return the number of lines of text we need to fit on the canvas.  This
   *  is used to set the height of the canvas.
   **/
  private int getLineCount () {
    if (show_labels) {
      return 17;
    } else {
      return 9;
    }
  }

  /**
   *  Return the vertical offset from the top of the canvas for this feature.
   *  The returned value will be the y-coordinate of the top of the lane that
   *  this feature should be draw into.
   **/
  private int getSegmentVerticalOffset (FeatureSegment segment) {
    // one "line" is font_height pixels high, (the number of lines times the
    // font_height is the height of the height of canvas)
    final int line = getSegmentDisplayLine (segment);

    return getLineOffset (line);
  }

  /**
   *  Given a line/frame ID (eg FORWARD_FRAME_1, BACKWARD_STRAND: etc.) return
   *  the vertical offset from the top of the canvas.
   **/
  private int getLineOffset (int line) {
    return line * getLineHeight ();
  }

  /**
   *  Return the lowest on screen (with respect to the canvas) x coordinate of
   *  a base.  If the scale_factor is zero then one base will be font_width
   *  wide and this method will return a different value than
   *  getHighXPositionOfBase ().  If scale_factor is greater than one then
   *  the two methods will return the same thing.
   *  @param base_number The (forward) base to calculate the position of.
   **/
  private int getLowXPositionOfBase (int base_number) {
    return (int) ((base_number - getFirstVisibleForwardBase ()) *
                  getScaleValue ());
  }

  /**
   *  Return the highest on screen (ie with respect to the canvas) x
   *  coordinate of a base.  See comment on getLowXPositionOfBase ().
   *  @param base_number The (forward) base to calculate the position of.
   **/
  private int getHighXPositionOfBase (int base_number) {
    if (getScaleFactor () == 0) {
      return getLowXPositionOfBase (base_number) + font_width - 1;
    } else {
      return getLowXPositionOfBase (base_number);
    }
  }

  /**
   *  Return the low on screen x coordinate of the base at the given Marker
   *  position.  If the scale_factor is zero then one base will be font_width
   *  wide so that this method will return a different position than
   *  getHighXPositionOfMarker ().
   *  @param marker The Marker position to return the screen position of.
   **/
  private int getLowXPositionOfMarker (Marker marker) {
    if (marker.getStrand ().isForwardStrand ()) {
      return getLowXPositionOfBase (marker.getRawPosition ());
    } else {
      return getHighXPositionOfBase (marker.getRawPosition ());
    }
  }

  /**
   *  Return the high on screen x coordinate of the base at the given Marker
   *  position.  If the scale_factor is zero then one base will be font_width
   *  wide so that this method will return a different position than
   *  getLowXPositionOfMarker ().
   *  @param marker The Marker position to return the screen position of.
   **/
  private int getHighXPositionOfMarker (Marker marker) {
    if (marker.getStrand ().isForwardStrand ()) {
      return getHighXPositionOfBase (marker.getRawPosition ());
    } else {
      return getLowXPositionOfBase (marker.getRawPosition ());
    }
  }


  /**
   *  This is the approximate opposite of getLowXPositionOfBase () - it returns
   *  MarkerRange corresponding to the given screen position.  It returns a
   *  MarkerRange rather than a Marker because if the user clicks on a FRAME
   *  line then we want to select a whole codon not just one base.
   *  @param position A position on the canvas.
   *  @return A MarkerRange covering the clicked on bases.
   **/
  private MarkerRange getMarkerRangeFromPosition (Point position) {
    return getMarkerRangeFromPosition (position, true);
  }

  /**
   *  This is the approximate opposite of getLowXPositionOfBase () - it returns
   *  MarkerRange corresponding to the given screen position.  It returns a
   *  MarkerRange rather than a Marker because if the user clicks on a FRAME
   *  line then we want to select a whole codon not just one base.
   *  @param position A position on the canvas.
   *  @param whole_codon If true then return a range that covers a whole codon
   *    if the click was on a frame line.
   *  @return A MarkerRange covering the clicked on bases.
   **/
  private MarkerRange getMarkerRangeFromPosition (Point position,
                                                  boolean whole_codon) {
    final int frame_id = getFrameFromPoint (position);

    int base_position;
    final Strand strand;

    // calculate the frame/strand line and base position
    switch (frame_id) {
    case FORWARD_FRAME_1:
    case FORWARD_FRAME_2:
    case FORWARD_FRAME_3:
    case FORWARD_STRAND:
      strand = getEntryGroup ().getBases ().getForwardStrand ();
      base_position = (int) (1.0 * position.x / getScaleValue () +
                             getFirstVisibleForwardBase ());
      break;
    case BACKWARD_FRAME_1:
    case BACKWARD_FRAME_2:
    case BACKWARD_FRAME_3:
    case BACKWARD_STRAND:
      {
        strand = getEntryGroup ().getBases ().getBackwardStrand ();

        final int raw_base_position =
          (int) (1.0 * position.x / getScaleValue () +
                 getFirstVisibleForwardBase ());

        base_position =
          strand.getBases ().getComplementPosition (raw_base_position);
      }
      break;
    default:
      base_position = 0;
      strand = null;
    }

    final int start_base_position;

    final int end_base_position;

    if (whole_codon) {
      start_base_position =
        adjustBasePositionForFrame (frame_id, base_position, true);
      end_base_position =
        adjustBasePositionForFrame (frame_id, base_position, false);
    } else {
      start_base_position = base_position;
      end_base_position = base_position;
    }

    if (strand == null) {
      return null;
    } else {
      try {
        return strand.makeMarkerRangeFromPositions (start_base_position,
                                                    end_base_position);
      } catch (OutOfRangeException e) {
        // XXX
        return null;
      }
    }
  }


  /**
   *  Helper method for getMarkerRangeFromPosition ().  Returns a base
   *  position offset to the start/end of the codon if the given frame id on a
   *  FRAME.
   *  @param get_start If true then the base position returned is base
   *    position of the start of the codon, otherwise the end base is
   *    returned.
   **/
  private int adjustBasePositionForFrame (final int frame_id,
                                          final int base_position,
                                          final boolean get_start) {
    final int return_base_position;

    final int end_offset;

    if (get_start) {
      end_offset = 0;
    } else {
      // the end is two bases ahead of the start
      end_offset = 2;
    }

    switch (frame_id) {
    case FORWARD_FRAME_1:
      {
        final int base_pos_mod3 = (base_position - 1) % 3;
        return_base_position = base_position + end_offset - base_pos_mod3;
      }
      break;
    case FORWARD_FRAME_2:
      {
        final int base_pos_mod3 = (base_position - 2) % 3;
        return_base_position = base_position + end_offset - base_pos_mod3;
      }
      break;
    case FORWARD_FRAME_3:
      {
        final int base_pos_mod3 = (base_position - 3) % 3;
        return_base_position = base_position + end_offset - base_pos_mod3;
      }
      break;
    case BACKWARD_FRAME_1:
      {
        final int base_pos_mod3 = (base_position - 1) % 3;
        return_base_position = base_position + end_offset - base_pos_mod3;
      }
      break;
    case BACKWARD_FRAME_2:
      {
        final int base_pos_mod3 = (base_position - 2) % 3;
        return_base_position = base_position + end_offset - base_pos_mod3;
      }
      break;
    case BACKWARD_FRAME_3:
      {
        final int base_pos_mod3 = (base_position - 3) % 3;
        return_base_position = base_position + end_offset - base_pos_mod3;
      }
      break;
    default:
      return_base_position = base_position;
      // do nothing
    }

    return return_base_position;
  }

  /**
   *  Draw one feature at the correct place in a Graphics object.
   *  @param g The Graphics object on which to draw.
   *  @param draw_feature_fill If true then draw the solid block of colour
   *    inside the segments.
   *  @param draw_feature_outline If true then the feature outline is
   *    drawn.
   *  @param feature The feature to draw
   **/
  private void drawFeature (Graphics g,
                            Feature feature,
                            boolean draw_feature_fill,
                            boolean draw_feature_outline) {
//     if (show_labels) {
//       System.out.println ("in drawFeature ()");
//     }

    if (feature.getKey ().equals ("source")) {
      return;
    }

    final FeatureSegmentVector this_feature_segments = feature.getSegments ();

    if (this_feature_segments.size () == 0) {
      // don't try to draw a feature with no segments
      return;
    }

    if (show_labels) {
      // draw the label first if there is a label line so that we don't
      // overwrite the edge of a feature
      drawFeatureLabel (g, feature);
    }

//     if (show_labels) {
//       System.out.println ("later in drawFeature ()");
//    }

    final Selection selection = getSelection ();
    final FeatureVector selection_features = selection.getSelectedFeatures ();

    // set to true if the whole of the current feature should be highlighted
    boolean highlight_feature_flag = false;

    // set to a vector of those segments of this feature that should be
    // highlighted
    FeatureSegmentVector highlight_segments = selection.getSelectedSegments ();

    if (selection_features.contains (feature)) {
      // ignore the possibility that a feature and a segment from the same
      // feature could be in the selection vector at the same time
      highlight_feature_flag = true;
    } else {
      // this feature isn't in the selection but maybe one of it's segments is.
      for (int i = 0 ; i < this_feature_segments.size () ; ++i) {
        final FeatureSegment current_segment =
          this_feature_segments.elementAt (i);

        if (highlight_segments.indexOf (current_segment) != -1) {
          // we highlight the segment and the feature
          highlight_feature_flag = true;
        }
      }
    }

    if (getSelection ().getSelectedFeatures ().size () == 1) {
//       System.out.println ("getSelection () size: " +
//                           getSelection ().getFeatures ().elementAt (0) + " " +
//                           feature);
//       System.out.println ("drawFeature (): " + highlight_feature_flag);
    }

//    System.out.println ("segments: " + this_feature_segments.size ());


    // draw each segment/exon
    for (int i = 0 ; i < this_feature_segments.size () ; ++i) {
      final FeatureSegment current_segment =
        this_feature_segments.elementAt (i);
      final boolean highlight_segment_flag;

      if (highlight_segments.indexOf (current_segment) == -1) {
        highlight_segment_flag = false;
      } else {
        highlight_segment_flag = true;
      }

      final boolean draw_direction_arrow_flag;

      // draw an arrow only on the last segment
      if (i == this_feature_segments.size () - 1) {
        draw_direction_arrow_flag = true;
      } else {
        draw_direction_arrow_flag = false;
      }

      drawSegment (g, current_segment,
                   highlight_feature_flag, highlight_segment_flag,
                   draw_feature_fill,
                   draw_feature_outline,
                   draw_direction_arrow_flag);

      if (feature.isProteinFeature () &&
          i + 1 < this_feature_segments.size ()) {
          // draw a line between the segments

        final FeatureSegment next_segment =
          this_feature_segments.elementAt (i + 1);

        drawSegmentConnection (g, current_segment, next_segment);
      }
    }

    if (!show_labels) {
      // draw the label last if the is no label line because in this case the
      // label is draw on top of the feature segments
      drawFeatureLabel (g, feature);
    }

    Thread.yield ();
  }

  /**
   *  Draw a bent line between two segments which represents the connection
   *  between exons in feature.
   *  @param g The Graphics object on which to draw.
   *  @param lower_segment The reference of the segment that is closest to the
   *    start of the Strand.  The connection line will start at the beginning
   *    of this segment.
   *  @param upper_segment The reference of the segment that is furthest from
   *    the start of the Strand.  The connection line will finish at the end
   *    of this segment.
   **/
  private void drawSegmentConnection (Graphics g,
                                      FeatureSegment lower_segment,
                                      FeatureSegment upper_segment) {
    final Marker upper_segment_start_marker = upper_segment.getStart ();
    final Marker lower_segment_end_marker = lower_segment.getEnd ();

    int next_segment_start_coord =
      getLowXPositionOfMarker (upper_segment_start_marker);

    if (next_segment_start_coord > 16000) {
      // make sure we don't wrap around when drawing
      next_segment_start_coord = 16000;
    }

    if (next_segment_start_coord < -16000) {
      // make sure we don't wrap around when drawing
      next_segment_start_coord = -16000;
    }

    int this_segment_end_coord =
      getHighXPositionOfMarker (lower_segment_end_marker);

    if (this_segment_end_coord > 16000) {
      // make sure we don't wrap around when drawing
      this_segment_end_coord = 16000;
    }

    if (this_segment_end_coord < -16000) {
      // make sure we don't wrap around when drawing
      this_segment_end_coord = -16000;
    }

    final int this_segment_vertical_offset =
      getSegmentVerticalOffset (lower_segment);
    final int next_segment_vertical_offset =
      getSegmentVerticalOffset (upper_segment);

    // we draw the line with a bend in the middle - this is the vertical
    // position of the bend
    final int line_y_position_for_centre;

    if (this_segment_vertical_offset < next_segment_vertical_offset) {
      line_y_position_for_centre = this_segment_vertical_offset;
    } else {
      line_y_position_for_centre = next_segment_vertical_offset;
    }

    final int line_y_position_for_this =
      this_segment_vertical_offset + getFeatureHeight () / 2;
    final int line_y_position_for_next =
      next_segment_vertical_offset + getFeatureHeight () / 2;

    final int horizontal_position_of_centre =
      (this_segment_end_coord + next_segment_start_coord) / 2;

//     System.out.println ("drawSegmentConnection (): from: " +
//                         this_segment_end_coord + " to: " +
//                         next_segment_start_coord);

    final Feature segment_feature = lower_segment.getFeature ();
    final Color feature_colour = segment_feature.getColour ();

    if (feature_colour == null ||
        getSelection ().contains (lower_segment.getFeature ())) {
      // draw in black if no colour is specified or if the feature is selected
      g.setColor (Color.black);
    } else {
      g.setColor (feature_colour);
    }

    g.drawLine (this_segment_end_coord,
                line_y_position_for_this,
                horizontal_position_of_centre,
                line_y_position_for_centre);

    g.drawLine (horizontal_position_of_centre,
                line_y_position_for_centre,
                next_segment_start_coord,
                line_y_position_for_next);
  }

  /**
   *  Draw a label a feature.  This is a helper function for drawFeature ().
   *  If show_labels is true the labels will be drawn below the features,
   *  otherwise they will be drawn within the features.
   *  @param g The Graphics object on which to draw.
   *  @param feature The feature to draw the label for.
   **/
  private void drawFeatureLabel (Graphics g, Feature feature) {

    if (!show_labels && getScaleFactor () == 0) {
      // the three frame translation is visible when the scale factor is 0,
      // don't draw labels over it
      return;
    }

    String label_or_gene = feature.getIDString ();

    final String label = feature.getLabel ();

    if (label != null && label.equals ("*")) {
      // special case - don't display a label if the label qualifier is "*"
      label_or_gene = "";
    }

    final FontMetrics fm = g.getFontMetrics ();
    final int string_width = fm.stringWidth (label_or_gene);

    final FeatureSegment first_segment = feature.getSegments ().elementAt (0);
    final int label_x_coord;

    if (feature.isForwardFeature ()) {
      final int segment_start_pos =
        first_segment.getStart ().getRawPosition ();
      label_x_coord = getLowXPositionOfBase (segment_start_pos);
    } else {
      final int segment_end_pos =
        first_segment.getEnd ().getRawPosition ();
      // reverse the start and end on the complementary strand
      label_x_coord = getLowXPositionOfBase (segment_end_pos);
    }

    if (label_x_coord >= canvas.getSize ().width) {
      return;
    }

    if (label_x_coord + string_width <= 0) {
      // don't draw the label if it is not visible on screen
    }


    int vertical_offset =
      getSegmentVerticalOffset (first_segment);

    if (show_labels) {
      vertical_offset += getLineHeight (); // move to the label line
    }

    // save this so we can restore it later
    final Shape saved_clip = g.getClip ();

    if (show_labels) {
      // if we have a labels line draw a white background behind the
      // label

      g.setColor (Color.white);

      g.fillRect (label_x_coord - getFontWidth (),
                  vertical_offset,
                  string_width + getFontWidth () * 2,
                  getLineHeight ());
    } else {
      // if there is no label line clip to the size of the first segment
      // and draw in there
      final int segment_start_coord =
        getSegmentStartCoord (first_segment);
      final int segment_end_coord =
        getSegmentEndCoord (first_segment);

      if (segment_end_coord > segment_start_coord) {
        g.setClip (segment_start_coord, vertical_offset,
                   segment_end_coord - segment_start_coord,
                   getFontHeight ());
      } else {
        g.setClip (segment_end_coord, vertical_offset,
                   segment_start_coord - segment_end_coord,
                   getFontHeight ());
      }

//       System.out.println ("setting clip to: " + segment_start_coord +
//                           " " +  vertical_offset + " " +
//                           (segment_end_coord - segment_start_coord) +
//                           " " + getFontHeight ());
    }

    g.setColor (Color.black);

    g.drawString (label_or_gene, label_x_coord + 1,
                  vertical_offset +
                  font_base_line);

    if (!show_labels) {
      g.setClip (saved_clip);
    }
  }

  /**
   *  Return the position on the canvas where this segment starts.  The
   *  returned value will be -1 if the position is off the left of the screen
   *  and will be (width of the canvas) if the position is off the right of
   *  the screen.
   **/
  private int getSegmentStartCoord (FeatureSegment segment) {
    final Marker segment_start_marker = segment.getStart ();
    int segment_start_coord =
      getLowXPositionOfMarker (segment_start_marker);

    if (segment_start_coord > canvas.getSize ().width) {
      // make sure we don't wrap around when drawing
      segment_start_coord = canvas.getSize ().width;
    }

    if (segment_start_coord < 0) {
      // make sure we don't wrap around when drawing
      segment_start_coord = -1;
    }

    return segment_start_coord;
  }

  /**
   *  Return the position on the canvas where this segment ends.  The returned
   *  value will be -1 if the position is off the left of the screen and will
   *  be (width of the canvas) if the position is off the right of the screen.
   **/
  private int getSegmentEndCoord (FeatureSegment segment) {
    final Marker segment_end_marker = segment.getEnd ();
    int segment_end_coord = getHighXPositionOfMarker (segment_end_marker);

    if (segment_end_coord > canvas.getSize ().width) {
      // make sure we don't wrap around when drawing
      segment_end_coord = canvas.getSize ().width;
    }

    if (segment_end_coord < 0) {
      // make sure we don't wrap around when drawing
      segment_end_coord = -1;
    }

    return segment_end_coord;
  }

  /**
   *  Return true if the base given by the Marker is visible.
   **/
  private boolean baseVisible (Marker marker) {
    final int first_visible_base = getFirstVisibleForwardBase ();
    final int last_visible_base = getLastVisibleForwardBase ();

    final int marker_low_coord =
      getLowXPositionOfBase (marker.getRawPosition ());
    final int marker_high_coord =
      getHighXPositionOfBase (marker.getRawPosition ());

    if (marker.getRawPosition () < first_visible_base ||
        marker.getRawPosition () > last_visible_base) {
      return false;
    } else {
      return true;
    }
  }

  /**
   *  Return if and only if the segment is (partly) within the range of bases
   *  we are currently displaying.
   *  @param segment The FeatureSegment to check.
   **/
  private boolean segmentVisible (FeatureSegment segment) {
    int segment_start_coord = getSegmentStartCoord (segment);
    int segment_end_coord = getSegmentEndCoord (segment);

    if (segment_end_coord < 0 && segment_start_coord < 0 ||
        segment_start_coord >= canvas.getSize ().width &&
        segment_end_coord >= canvas.getSize ().width) {
      return false;
    } else {
      return true;
    }
  }

  /**
   *  Return if and only if some part of a feature is (partly) within the
   *  range of bases we are currently displaying.
   *  @param feature The Feature to check.
   **/
  private boolean featureVisible (Feature feature) {
    if (getMinimumScore () > 0) {
      final int feature_score = feature.getScore ();

      if (feature_score != -1 && feature_score < getMinimumScore ()) {
        return false;
      }
    }

    final FeatureSegmentVector segments = feature.getSegments ();

    for (int i = 0 ; i < segments.size () ; ++i) {
      if (segmentVisible (segments.elementAt (i))) {
        return true;
      }
    }

    return false;
  }

  /**
   *  Draw one FeatureSegment into a Graphics object.
   *  @param g The Graphics object on which to draw.
   *  @param segment The feature segment to draw
   *  @param highlight_feature If true draw an extra thick line
   *  @param highlight_segment If true draw the segment with a doubly thick
   *    line.
   *  @param draw_feature_fill If true then draw the solid block of colour
   *    inside the segments.
   *  @param draw_feature_outline If true then the feature outline is
   *    drawn.
   *  @param draw_feature_outline If true then only the feature outline is
   *    drawn, if false then only the solid block of colour is drawn.
   *  @param draw_arrow If true draw a direction arrow at the end of the
   *    segment.
   **/
  private void drawSegment (Graphics g, FeatureSegment segment,
                            boolean highlight_feature,
                            boolean highlight_segment,
                            boolean draw_feature_fill,
                            boolean draw_feature_outline,
                            boolean draw_arrow) {
    if (!segmentVisible(segment)) {
      // not on screen
      return;
    }

    final Feature segment_feature = segment.getFeature ();

    final int vertical_offset = getSegmentVerticalOffset (segment) + 1;

    int segment_start_coord = getSegmentStartCoord (segment);
    int segment_end_coord = getSegmentEndCoord (segment);

//       System.out.println ("drawSegment (): before - " +
//                           segment.getFeature ().getLabel () + " " +
//                           segment_start_coord + " " +
//                           segment_end_coord + " " +
//                           getFirstVisibleForwardBase ());

    final int segment_width  = segment_end_coord - segment_start_coord + 1;

//     System.out.println ("scroll value: " + scrollbar.   getBlockIncrement() );
//     System.out.println ("scroll value: " + scrollbar.   getUnitIncrement() );
//     System.out.println ("scroll value: " + scrollbar.   getMaximum() );
//     System.out.println ("scroll value: " + scrollbar.   getMinimum() );
//     System.out.println ("scroll value: " + scrollbar.   getOrientation() );
//     System.out.println ("scroll value: " + scrollbar.   getValue() );
//     System.out.println ("scroll value: " + scrollbar.   getVisibleAmount());

//     System.out.println ("drawSegment (): after" +
//                         segment.getFeature ().getLabel () + " " +
//                         segment_start_coord + " " +
//                         segment_end_coord);


    final int segment_height = getFeatureHeight () - 1;

//     System.out.println ("drawSegment (): " +
//                        segment.getFeature ().getLabel ());

    // this is 1 if the feature is on the forward strand or on a forward frame
    // and -1 otherwise.  this used to draw the feature arrow in the right
    // direction.
    final int feature_direction;
    if (segment.getFeature ().isForwardFeature ()) {
      feature_direction = 1;
    } else {
      feature_direction = -1;
    }

    final int [] x_points = {
      segment_start_coord, segment_end_coord,
      segment_end_coord, segment_start_coord
    };

    final int [] y_points = {
      vertical_offset, vertical_offset,
      vertical_offset + segment_height, vertical_offset + segment_height
    };

    final Color feature_colour = segment_feature.getColour ();

    if (feature_colour == null) {
      // no colour means draw in white
      g.setColor (Color.white);
    } else {
      g.setColor (feature_colour);
    }

    if (draw_feature_fill) {
      g.fillPolygon (x_points, y_points, 4);
    }

    if (draw_feature_outline) {
      g.setColor (Color.black);
      g.drawPolygon (x_points, y_points, 4);

      if (highlight_feature) {
        // selected - highlight by drawing a thicker line

        x_points[0] -= feature_direction; x_points[1] += feature_direction;
        x_points[2] += feature_direction; x_points[3] -= feature_direction;
        y_points[0] -= 1; y_points[1] -= 1;
        y_points[2] += 1; y_points[3] += 1;
        g.drawPolygon (x_points, y_points, 4);

        x_points[0] -= feature_direction; x_points[1] += feature_direction;
        x_points[2] += feature_direction; x_points[3] -= feature_direction;
        y_points[0] -= 1; y_points[1] -= 1;
        y_points[2] += 1; y_points[3] += 1;
        g.drawPolygon (x_points, y_points, 4);

        if (highlight_segment) {
          x_points[0] -= feature_direction; x_points[1] += feature_direction;
          x_points[2] += feature_direction; x_points[3] -= feature_direction;
          y_points[0] -= 1; y_points[1] -= 1;
          y_points[2] += 1; y_points[3] += 1;
          g.drawPolygon (x_points, y_points, 4);

          x_points[0] -= feature_direction; x_points[1] += feature_direction;
          x_points[2] += feature_direction; x_points[3] -= feature_direction;
          y_points[0] -= 1; y_points[1] -= 1;
          y_points[2] += 1; y_points[3] += 1;
          g.drawPolygon (x_points, y_points, 4);
        }
      } else {
        // not selected - don't highlight
      }

      if (draw_arrow) {
        // now draw the arrow point
        final int arrow_tip_x =
          x_points[1] + feature_direction * getFontWidth () * 8 / 10;
        final int arrow_tip_y = (y_points[1] + y_points[2]) / 2;

        g.drawLine (x_points[1], y_points[1], arrow_tip_x, arrow_tip_y);
        g.drawLine (arrow_tip_x, arrow_tip_y, x_points[2], y_points[2]);
      }
    }
  }

  /**
   *  Draw/highlight the selected range of bases on the FORWARD_STRAND or
   *  BACKWARD_STRAND lines.
   **/
  private void drawBaseSelection (Graphics g) {
    final MarkerRange selection_range =
      getSelection ().getMarkerRange ();

    if (selection_range == null) {
      return;
    }

    final Marker raw_start_coord;
    final Marker raw_end_coord;

    if (selection_range.getStrand ().isForwardStrand ()) {
      raw_start_coord = selection_range.getRawStart ();
      raw_end_coord = selection_range.getRawEnd ();
    } else {
      raw_start_coord = selection_range.getRawEnd ();
      raw_end_coord = selection_range.getRawStart ();
    }

    final int start_coord = getLowXPositionOfMarker (raw_start_coord);
    final int end_coord = getHighXPositionOfMarker (raw_end_coord);

    if (start_coord > canvas.getSize ().width &&
        end_coord > canvas.getSize ().width) {
      // not on screen
      return;
    }

    if (start_coord < 0 && end_coord < 0) {
      // not on screen
      return;
    }

    // the ID of the strand on which to draw the highlighting - this will be
    // FORWARD_STRAND or BACKWARD_STRAND.
    final int strand_id;

    // the ID of the strand on which to draw the highlighting - this will be
    // FORWARD_FRAME_1, 2 or 3 or BACKWARD_FRAME_1, 2 or 3.
    final int frame_id;

    if (selection_range.getStrand ().isForwardStrand ()) {
      strand_id = FORWARD_STRAND;
      switch ((selection_range.getStart ().getPosition () - 1) % 3) {
      case 0:
        frame_id = FORWARD_FRAME_1;
        break;
      case 1:
        frame_id = FORWARD_FRAME_2;
        break;
      case 2:
        frame_id = FORWARD_FRAME_3;
        break;
      default:
        frame_id = NO_FRAME;
      }
    } else {
      strand_id = BACKWARD_STRAND;
      switch ((selection_range.getStart ().getPosition () - 1) % 3) {
      case 0:
        frame_id = BACKWARD_FRAME_1;
        break;
      case 1:
        frame_id = BACKWARD_FRAME_2;
        break;
      case 2:
        frame_id = BACKWARD_FRAME_3;
        break;
      default:
        frame_id = NO_FRAME;
      }
    }

    drawBaseRange (g, start_coord, end_coord, strand_id, Color.yellow);
    drawBaseRange (g, start_coord, end_coord, frame_id, Color.pink);
  }

  /**
   *  Draw a rectangle representing the currently selected bases.
   *  @param g The Graphics object on which to draw.
   *  @param start_coord The start x coordinate.
   *  @param start_coord The end x coordinate.
   *  @param frame_id The ID of the frame line where the box should be drawn.
   *  @param fill_colour The colour to use to draw the box (the border will be
   *    black).
   **/
  private void drawBaseRange (Graphics g,
                             int start_coord, int end_coord,
                             int frame_id,
                             Color fill_colour) {
    if (start_coord < -1) {
      start_coord = -1;
    }

    final int canvas_width = canvas.getSize ().width;

    if (end_coord > canvas_width) {
      end_coord = canvas_width;
    }

    final int frame_line = getFrameDisplayLine (frame_id);

    final int frame_line_y_coord = getLineOffset (frame_line);

    final int [] x_points = {
      start_coord, end_coord, end_coord, start_coord
    };

    final int [] y_points = {
      frame_line_y_coord, frame_line_y_coord,
      frame_line_y_coord + getFeatureHeight () + 1,
      frame_line_y_coord + getFeatureHeight () + 1
    };

    g.setColor (fill_colour);

    g.fillPolygon (x_points, y_points, 4);

    g.setColor (Color.black);

    g.drawPolygon (x_points, y_points, 4);

  }

  /**
   *  Create and add the map_panel and it's contents.
   **/
  private void createMapPanel (final boolean scrollbar_at_top) {
    map_panel = new Panel ();

    map_panel.setLayout (new BorderLayout());

    createCanvas ();

    createScrollbar (scrollbar_at_top);

    add (map_panel, "Center");
  }

  /**
   *  Create the canvas object for this FeatureDisplay and add it to map_panel.
   **/
  private void createCanvas () {
    canvas = new Canvas ()
      {
      /**
       *  Set the offscreen buffer to null as part of invalidation.
       **/
      public void invalidate() {
        super.invalidate();
        offscreen = null;
      }

      /**
       *  Override update to *not* erase the background before painting
       */
      public void update(Graphics g) {
        paint(g);
      }

      /**
       *  Paint the canvas.
       */
      public void paint(Graphics g) {
        paintCanvas (g);
      }

    };

    fixDisplay ();
    map_panel.add (canvas, "Center");

    canvas.addMouseListener (new MouseAdapter () {
      /**
       *  Listen for mouse press events so that we can do popup menus and
       *  selection.
       **/
      public void mousePressed (MouseEvent event) {
//        System.out.println ("mousePressed ()");

        handleCanvasMousePress (event);

        if (event.isPopupTrigger () || event.isMetaDown ()) {
          final FeaturePopup popup =
            new FeaturePopup (FeatureDisplay.this,
                              getEntryGroup (),
                              getSelection (),
                              getGotoEventSource ());
          final Canvas parent = (Canvas) event.getSource ();

//          System.out.println ("mousePressed () - popup");

          parent.add (popup);

          popup.show (parent, event.getX (), event.getY ());
        }
      }
    });

    /**
     *  Listen for mouse motion events so that we can select ranges of bases.
     **/
    canvas.addMouseMotionListener (new MouseMotionAdapter () {
      public void mouseDragged (MouseEvent event) {
        handleCanvasMouseDrag (event);
      }
    });
  }

  /**
   *  Handle a mouse press event on the drawing canvas - select on click,
   *  select and broadcast it on double click.
   **/
  private void handleCanvasMousePress (MouseEvent event) {
//    System.out.println ("handleCanvasMousePress ()");
    if (event.getID() != MouseEvent.MOUSE_PRESSED) {
      return;
    }

    if (event.getClickCount () == 2) {
      handleCanvasDoubleClick (event);
    } else {
      handleCanvasSingleClick (event);
    }

    repaintCanvas ();
  }

  /**
   *  Handle a double click on the canvas.
   **/
  private void handleCanvasDoubleClick (MouseEvent event) {
    if (event.isPopupTrigger () || event.isMetaDown ()) {
      return;
    }

    if ((event.getModifiers () & InputEvent.BUTTON2_MASK) != 0 ||
        event.isAltDown () || event.isControlDown ()) {

      final Selectable clicked_thing = getThingAtPoint (event.getPoint ());

      if (clicked_thing == null) {
        // select the ORF around the click position
        final MarkerRange new_click_range =
          getMarkerRangeFromPosition (event.getPoint ());

        if (new_click_range == null) {
          return;
        }

        final Bases bases = getEntryGroup ().getBases ();

        final MarkerRange orf_range =
          Strand.getORFAroundMarker (new_click_range.getStart ());

        if (orf_range == null) {
          // user clicked on a stop codon
          return;
        } else {
          getSelection ().setMarkerRange (orf_range);
        }
      } else {
        // edit the selected Feature

        final Feature clicked_feature = getFeatureOf (clicked_thing);

        getSelection ().set (clicked_feature);

        if (Diana.isStandAlone ()) {
          new FeatureEdit (clicked_feature, getSelection ()).show ();
        }
      }
    }

    makeSelectionVisible ();
  }

  /**
   *  Handle a single click on the canvas.
   **/
  private void handleCanvasSingleClick (MouseEvent event) {

    click_segment_marker = null;

    final Selectable clicked_thing = getThingAtPoint (event.getPoint ());

    // any modifier apart from shift and control is treated like pressing
    // button 2
    if (clicked_thing == null ||
        (event.getModifiers () & InputEvent.BUTTON2_MASK) != 0 ||
        event.isAltDown () || event.isControlDown ()) {

      // if the user presses the mouse button 2 on feature or segment we treat
      // it like the feature/segment isn't there

      if (event.isPopupTrigger () || event.isMetaDown ()) {
        return;
      }

      if (event.isShiftDown () &&
          getSelection ().getAllFeatures ().size () > 0) {
        // something is selected but the user clicked on nothing with the
        // shift key down - do nothing
        return;
      }

      final MarkerRange old_selected_range =
        getSelection ().getMarkerRange ();

      final MarkerRange new_click_range =
        getMarkerRangeFromPosition (event.getPoint ());

      if (new_click_range == null) {
        return;
      }

      final MarkerRange new_selection_range;

      if (!event.isShiftDown () || old_selected_range == null) {
        new_selection_range = new_click_range;
      } else {
        if (old_selected_range.getStrand () ==
            new_click_range.getStrand ()) {
          new_selection_range =
            old_selected_range.combineRanges (new_click_range, true);
        } else {
          new_selection_range = old_selected_range;
        }
      }

      getSelection ().setMarkerRange (new_selection_range);

      click_range = new_selection_range;
    } else {

      if (getSelection ().getMarkerRange () != null &&
          event.isShiftDown ()) {
        return;
      }

      // clear the MarkerRange because this isn't the start of a range drag
      click_range = null;

      getSelection ().setMarkerRange (null);

      final Feature clicked_feature = getFeatureOf (clicked_thing);

      raiseFeature (clicked_feature);

      if (clicked_thing instanceof Feature) {
        // toggle the feature in or out of the selection unless this event
        // is a popup trigger in which case we should just make sure the
        // feature is in the selection
        if (getSelection ().contains (clicked_feature)) {
          if (! (event.isPopupTrigger () || event.isMetaDown ())) {
            if (event.isShiftDown ()) {
              getSelection ().remove (clicked_feature);
            } else {
              getSelection ().set (clicked_feature);
            }
          }
        } else {
          if (event.isShiftDown ()) {
            getSelection ().add (clicked_feature);
          } else {
            getSelection ().set (clicked_feature);
          }
        }
      } else {
        // must be a FeatureSegment

        // toggle the feature segment in or out of the selection unless
        // this event is a popup trigger in which case we should just make
        // sure the segment is in the selection
        final FeatureSegment clicked_segment =
          (FeatureSegment) clicked_thing;

        final FeatureSegmentVector selected_feature_segments =
          getSelection ().getSelectedSegments ();

        if (selected_feature_segments.contains (clicked_segment)) {
          if (! (event.isPopupTrigger () || event.isMetaDown ())) {
            if (event.isShiftDown ()) {
              getSelection ().remove (clicked_segment);
            } else {
              getSelection ().set (clicked_segment);
            }
          }
        } else {
          if (event.isShiftDown ()) {
            final FeatureVector selected_features =
              getSelection ().getSelectedFeatures ();

            if (selected_features.contains (clicked_feature)) {
              getSelection ().remove (clicked_feature);
            } else {
              getSelection ().add (clicked_segment);
            }
          } else {
            getSelection ().set (clicked_segment);
          }
        }
      }

      if (Options.getOptions ().canDirectEdit ()) {
        // click to see if the click was on a marker, the second argument is
        // false because we want the range to cover just one base
        final MarkerRange new_click_range =
          getMarkerRangeFromPosition (event.getPoint (), false);

        // search the feature to find if the start Marker of any of the
        // segments match the start Marker of new_click_range or if an end
        // Marker matches the end Marker of new_click_range

        final FeatureSegmentVector segments = clicked_feature.getSegments ();

        for (int i = 0 ; i < segments.size () ; ++i) {
          final FeatureSegment this_segment = segments.elementAt (i);

          if (new_click_range.getStart ().equals (this_segment.getStart ()) &&
              this_segment.canDirectEdit ()) {
            click_segment_marker = this_segment.getStart ();
            click_segment_marker_is_start_marker = true;
            other_end_of_segment_marker = this_segment.getEnd ();
            getSelection ().setMarkerRange (new_click_range);
            break;
          }

          if (new_click_range.getEnd ().equals (this_segment.getEnd ()) &&
              this_segment.canDirectEdit ()) {
            click_segment_marker = this_segment.getEnd ();
            click_segment_marker_is_start_marker = false;
            other_end_of_segment_marker = this_segment.getStart ();
            getSelection ().setMarkerRange (new_click_range);
            break;
          }
        }
      }

//      System.out.println ("marker: " + click_segment_marker);
    }
  }


  /**
   *  This is a helper method for handleCanvasSingleClick () and
   *  handleCanvasDoubleClick ().
   *  @param object The should be an instance of Feature or FeatureSegment.
   *  @return If the the argument is a Feature it is returned, if it is a
   *    FeatureSegment the owning Feature is returned.
   **/
  private Feature getFeatureOf (final Object object) {
    if (object instanceof Feature) {
      return (Feature) object;
    } else {
      // object must be a FeatureSegment
      return ((FeatureSegment) object).getFeature ();
    }
  }

  /**
   *  Handle a mouse drag that started on a Marker from a FeatureSegment.
   **/
  private void handleMarkerDrag (final MouseEvent event) {
    MarkerRange drag_range =
      getMarkerRangeFromPosition (event.getPoint (), false);

    if (drag_range == null) {
      return;
    }

    final int new_position;

    if (click_segment_marker_is_start_marker) {
      // the Marker is at the start of the segment
      new_position = drag_range.getStart ().getPosition ();

      if (new_position > other_end_of_segment_marker.getPosition ()) {
        // don't go past the other end of the segment
        return;
      }
    } else {
      new_position = drag_range.getEnd ().getPosition ();

      if (new_position < other_end_of_segment_marker.getPosition ()) {
        // don't go past the other end of the segment
        return;
      }
    }

    try {
      click_segment_marker.setPosition (new_position);
    } catch (OutOfRangeException e) {
      throw new Error ("internal error - unexpected OutOfRangeException");
    }
  }

  /**
   *  Handle a mouse drag event on the drawing canvas.
   **/
  private void handleCanvasMouseDrag (final MouseEvent event) {
    if (event.isShiftDown () &&
        getSelection ().getAllFeatures ().size () > 0) {
      return;
    }

    if (click_segment_marker != null) {
      handleMarkerDrag (event);
      return;
    }

    MarkerRange drag_range = getMarkerRangeFromPosition (event.getPoint ());

    if (drag_range == null) {
      return;
    }

    final MarkerRange selected_range =
      getSelection ().getMarkerRange ();

    if (selected_range != null &&
        drag_range.getStrand () != selected_range.getStrand ()) {
      // if the start and current positions of the drag are not on the
      // same Strand then ignore this event
      return;
    }

    try {
      if (drag_range.getStart ().getPosition () < 1) {
        // user has dragged off the screen so use a marker at position 1
        drag_range = new MarkerRange (drag_range.getStrand (), 1, 1);
      }

      if (drag_range.getEnd ().getPosition () > getSequenceLength ()) {
        // user has dragged off the screen so use a marker at position the
        // end of the sequence
        drag_range = new MarkerRange (drag_range.getStrand (),
                                      getSequenceLength (),
                                      getSequenceLength ());
      }
    } catch (OutOfRangeException e) {
      throw new Error ("internal error - unexpected OutOfRangeException");
    }

    final boolean start_base_is_visible = baseVisible (drag_range.getStart ());

    final boolean end_base_is_visible = baseVisible (drag_range.getEnd ());

    if (!start_base_is_visible) {
      makeBaseVisible (drag_range.getStart ());
    }

    if (!end_base_is_visible) {
      makeBaseVisible (drag_range.getEnd ());
    }

    if (!start_base_is_visible || !end_base_is_visible) {
      // scrolling was necessary so update the visible features vector
      updateVisibleFeatureVector ();
    }

    final MarkerRange new_marker_range;

    if (click_range == null) {
      new_marker_range = drag_range;
    } else {
      new_marker_range = selected_range.combineRanges (drag_range, true);
    }

    getSelection ().setMarkerRange (new_marker_range);

    repaintCanvas ();
  }

  /**
   *  This method sends an GotoEvent to all the GotoEvent listeners that will
   *  make the start of the first selected Feature or FeatureSegment visible.
   **/
  private void makeSelectionVisible () {
    final GotoEvent new_event =
      new GotoEvent (this, getSelection ().getStartBaseOfSelection ());

    getGotoEventSource ().sendGotoEvent (new_event);
  }

  /**
   *  Return the reference of the Object at p on the canvas or null if there
   *  is none.  The returned Object will be a Feature or a FeatureSegment.
   **/
  private Selectable getThingAtPoint (Point p) {
    final int frame_of_point = getFrameFromPoint (p);
    if (frame_of_point == NO_FRAME) {
      // point is not on a forward or backward frame or strand
      return null;
    }

    // go through the features in reverse order because the feature that is
    // drawn last will be on top
    for (int i = getVisibleFeatures ().size () - 1 ; i >= 0 ; --i) {
      final Feature current_feature = getVisibleFeatures ().elementAt (i);
      final FeatureSegmentVector segments = current_feature.getSegments ();

      for (int segment_index = 0 ;
           segment_index < segments.size () ;
           ++segment_index) {
        final FeatureSegment current_segment =
          segments.elementAt (segment_index);

        if (current_segment.getFrameID (getFrameFeaturesFlag ()) ==
            frame_of_point) {
          // this segment is in the right frame so check that it is between
          // the start and end positions of the segment

          if (p.x >= getSegmentStartCoord (current_segment) &&
              p.x <= getSegmentEndCoord (current_segment) ||
              p.x <= getSegmentStartCoord (current_segment) &&
              p.x >= getSegmentEndCoord (current_segment)) {
            final Feature segment_feature = current_segment.getFeature ();

            if (segment_feature.getSegments ().size () == 1) {
              // special case - if there is only one segment then select the
              // whole feature
              return segment_feature;
            } else {
              return current_segment;
            }
          }
        }
      }
    }

    return null;
  }

  /**
   *  Create and add a new scroll bar to this FeatureDisplay.
   **/
  private void createScrollbar (final boolean scrollbar_at_top) {
    scrollbar = new Scrollbar (Scrollbar.HORIZONTAL);
//    scrollbar.setBackground (Color.green);
    scrollbar.addAdjustmentListener (new AdjustmentListener () {
      public void adjustmentValueChanged(AdjustmentEvent e) {
        if (first_visible_base != e.getValue ()) {
          first_visible_base = e.getValue ();
          fireAdjustmentEvent ();

          updateVisibleFeatureVector ();
          repaintCanvas ();
        }
      }
    });

    if (scrollbar_at_top) {
      map_panel.add (scrollbar, "North");
    } else {
      map_panel.add (scrollbar, "South");
    }
    fixScrollbar ();
  }

  /**
   *  Send a DisplayAdjustmentEvent with the current start and end base to
   *  each of the listeners.  This has package scope because EntryEdit
   *  components need to send this event when a new BasePlot component is
   *  added.  The BasePlot needs to know the first visible base, last visible
   *  base and the maximum number of visible bases before the plot can be
   *  drawn.
   **/
  void fireAdjustmentEvent () {
    final DisplayAdjustmentEvent event =
      new DisplayAdjustmentEvent (this,
                                  getFirstVisibleForwardBase (),
                                  getLastVisibleForwardBase (),
                                  getMaxVisibleBases (),
                                  getScaleValue (),
                                  getScaleFactor ());

    fireAction (adjustment_listener_list, event);
  }

  /**
   *  Send an event to those object listening for it.
   *  @param listeners A Vector of the objects that the event should be sent
   *    to.
   *  @param event The event to send
   **/
  private void fireAction (Vector listeners, ChangeEvent event) {
    final Vector targets;
    // copied from a book - synchronizing the whole method might cause a
    // deadlock
    synchronized (this) {
      targets = (Vector) listeners.clone ();
    }

    for ( int i = 0 ; i < targets.size () ; ++i ) {
      ChangeListener change_listener = (ChangeListener) targets.elementAt (i);

      final DisplayAdjustmentListener target =
        (DisplayAdjustmentListener) change_listener;
      target.displayAdjustmentValueChanged ((DisplayAdjustmentEvent) event);
    }
  }

  /**
   *  Create the scroll bar used for changing the scale and add it to the
   *  FeatureDisplay.
   **/
  private void createScaleScrollbar () {
    scale_changer = new Scrollbar (Scrollbar.VERTICAL);
    // try to arrange for the scrollbar to have a maximum value big enough
    // that the whole sequence can be visible at once
    final int MAX_FACTOR =
      (int) Math.round (Math.log (getSequenceLength ()/20) /  Math.log (3));
    scale_changer.setValues (getScaleFactor (), 1, 0, MAX_FACTOR);
    scale_changer.setBlockIncrement (1);
    scale_changer.setUnitIncrement (1);
    scale_changer.addAdjustmentListener (new AdjustmentListener () {
      public void adjustmentValueChanged(AdjustmentEvent e) {
        setScaleFactor (e.getValue ());
      }
    });

    add (scale_changer, "East");

    if (scale_factor >= MAX_FACTOR) {
      setScaleFactor (MAX_FACTOR - 1);
    }
  }

  /**
   *  Update the parameters of the scrollbar taking changes to the entry_group
   *  into account.
   **/
  private void fixScrollbar () {
    if (scrollbar == null) {
      return;
    }

    if (getEntryGroup ().size () == 0) {
      scrollbar.setValues (1, 255, 1, 255);
    } else {
      final int sequence_length = getSequenceLength ();

      final int max_visible_bases;

      // make sure max_visible_bases is at least 1 to stop setBlockIncrement ()
      // from complaining
      if (getMaxVisibleBases () > 0) {
        max_visible_bases = getMaxVisibleBases ();
      } else {
        max_visible_bases = 1;
      }

      // initial_value, visible, minimum, maximum
      scrollbar.setValues (getFirstVisibleForwardBase (),
                           max_visible_bases,
                           1,
                           sequence_length + max_visible_bases/2);

      scrollbar.setBlockIncrement (max_visible_bases);

      if (max_visible_bases >= 10 && getScaleFactor () > 0) {
        scrollbar.setUnitIncrement (max_visible_bases/10);
      } else {
        scrollbar.setUnitIncrement (1);
      }

      fireAdjustmentEvent ();
    }
  }

  /**
   *  Set the size of the canvas, taking the value of font_height and
   *  show_label into account.
   **/
  private void fixDisplay () {
    if (canvas == null) {
      return;
    }

    final int new_width = canvas.getSize ().width;
    final int new_height = getLineHeight () * getLineCount ();

    if (new_height != canvas.getSize ().width ||
        new_width != canvas.getSize ().height) {

      canvas.setSize (new_width, new_height);

      repaintCanvas ();
    }

    fixScrollbar ();
  }

  /**
   *  Returns the base length of the sequence we are displaying in this
   *  component.
   **/
  public int getSequenceLength () {
    return getEntryGroup ().getSequenceLength ();
  }

  /**
   *  Set the first visible base in the forward direction (set the base on the
   *  forward strand that is on the left of the canvas).  This method will
   *  scroll the display so that this base is is at the very left hand side of
   *  the canvas.
   **/
  private void setFirstVisibleForwardBase (int new_position) {
    if (first_visible_base != new_position) {
      first_visible_base = new_position;
      fixScrollbar ();
      updateVisibleFeatureVector ();
      repaintCanvas ();
    }
    //    System.out.println (new_position);
  }

  /**
   *  Set the last visible base in the forward direction (set the base on the
   *  forward strand that is on the right of the canvas).  This method will
   *  scroll the display so that this base is is at the very right hand side
   *  of the canvas (or as close to that position as possible.
   **/
  private void setLastVisibleForwardBase (int new_position) {
    final int possible_first_base = new_position - getMaxVisibleBases ();

    if (possible_first_base < 1) {
      setFirstVisibleForwardBase (1);
    } else {
      setFirstVisibleForwardBase (possible_first_base);
    }
  }

  /**
   *  Set the first visible base in the backward direction.  ie. set the base
   *  on the forward strand that is on the left of the canvas.  This method
   *  will scroll the display so that this base is is at the very left hand
   *  side of the canvas.
   **/
  private void setFirstVisibleBackwardBase (int new_position) {
    final Bases bases = getEntryGroup ().getBases ();
    final int new_last_forward_base =
      bases.getComplementPosition (new_position);

    setLastVisibleForwardBase (new_last_forward_base);
  }

  /**
   *  Call repaint () on the canvas object.
   **/
  private void repaintCanvas () {
//      try {
//        throw new Error ("");
//      } catch (Error e) {
//        System.out.println ("repaintCanvas ():");
//        e.printStackTrace ();
//      }

    canvas.repaint ();
  }

  /**
   *  Return the current default font (from Diana.options).
   **/
  private Font getDefaultFont () {
    return Options.getOptions ().getFont ();
  }

  /**
   *  Return the width of our font, as calculated by setFontInfo ().
   **/
  private int getFontWidth () {
    return font_width;
  }

  /**
   *  Return the height of our font, as calculated by setFontInfo ().
   **/
  private int getFontHeight () {
    return font_height;
  }

  /**
   *  Return the height each line of the display should be.  Each frame will
   *  be drawn into one line.
   **/
  private int getLineHeight () {
    return getFontHeight () + 2;
  }

  /**
   *  Return the current scale factor.  The scale factor is a number greater
   *  than or equal to zero than controls the number of bases that can appear
   *  on screen.  See getScaleValue ().
   **/
  public int getScaleFactor () {
    return scale_factor;
  }

  /**
   *  Return the ID of the frame at the given Point in the canvas or NO_FRAME
   *  if the Point is not within a frame. If labels are on then points on the
   *  label lines get returned as the frame id of the STRAND or FRAME line
   *  above.
   **/
  private int getFrameFromPoint (Point point) {
    // this is a hack - the id values go from 0 to 7 (FORWARD_STRAND to
    // BACKWARD_FRAME_1).  we test each in turn.
    for (int i = FORWARD_STRAND ;
         i <= BACKWARD_FRAME_1 ;
         ++i) {
      final int top_of_line = getLineOffset (getFrameDisplayLine (i));
      final int line_height;

      if (show_labels) {
        line_height = getLineHeight () * 2;
      } else {
        line_height = getLineHeight ();
      }

      if (point.y > top_of_line &&
          point.y < top_of_line + line_height) {
//        System.out.println ("getFrameFromPoint () - " + i);
        return i;
      }
    }
    // the point isn't in any or the frames
    return NO_FRAME;
  }

  /**
   *  Return the base number of the first visible base in the forward
   *  direction, ie the base on the forward strand that is on the left
   *  each of the canvas.
   **/
  public int getFirstVisibleForwardBase () {
    return first_visible_base;
  }

  /**
   *  Return a base marker of the first visible base in the forward
   *  direction, ie the base on the forward strand that is on the left
   *  each of the canvas.
   **/
  private Marker getFirstVisibleForwardBaseMarker () {
    try {
      final Strand strand = getEntryGroup ().getBases ().getForwardStrand ();
      return strand.makeMarker (getFirstVisibleForwardBase ());
    } catch (OutOfRangeException e) {
      throw new Error ("internal error - unexpected OutOfRangeException");
    }
  }

  /**
   *  Return the base number of the last visible base in the forward
   *  direction, ie the base on the forward strand that is on the right
   *  edge of the canvas.
   **/
  private int getLastVisibleForwardBase () {
    final int possible_last_base =
      (int) (getFirstVisibleForwardBase () + getMaxVisibleBases ());

    if (possible_last_base > getSequenceLength ()) {
      return getSequenceLength ();
    } else {
      return possible_last_base;
    }
  }

  /**
   *  Return the base number of the first visible base in the backward
   *  direction , ie the base on the backward strand that is on the left
   *  each of the canvas.
   **/
  private int getFirstVisibleBackwardBase () {
    final int last_forward_base = getLastVisibleForwardBase ();
    final Bases bases = getEntryGroup ().getBases ();
    return bases.getComplementPosition (last_forward_base);
  }

  /**
   *  Return the base number of the last visible base in the backward
   *  direction, ie the base on the backward strand that is on the right
   *  edge of the canvas.
   **/
  private int getLastVisibleBackwardBase () {
    final int first_forward_base = getFirstVisibleForwardBase ();
    final Bases bases = getEntryGroup ().getBases ();
    return bases.getComplementPosition (first_forward_base);
  }

  /**
   *  Return the base of the sequence that is closest to the centre of the
   *  view.
   **/
  private int getCentreForwardBase () {
    int possible_return_position =
      getFirstVisibleForwardBase () + getMaxVisibleBases () / 2;

    if (possible_return_position < getSequenceLength ()) {
      return possible_return_position;
    } else {
      return getSequenceLength ();
    }
  }

  /**
   *  Return the amount display is scaled.  The value is 3 to the power of
   *  (scale_factor - 1), except for a scale_factor of zero which gives a
   *  scale value of 1/font_width.
   **/
  private float getScaleValue () {
    return scale_value;
  }

  /**
   *  Return the height in pixels we should use for drawing features.
   **/
  private int getFeatureHeight () {
    // don't use getLineHeight () because we want a nice space between
    // frames/lines
    return getFontHeight ();
  }

  /**
   *  Return the number of bases we can fit on screen at once, ie the number
   *  that will fit side by side on the canvas.
   **/
  private int getMaxVisibleBases () {
    final int canvas_width = canvas.getSize ().width;

    return (int)(canvas_width / getScaleValue ());
  }

  /**
   *  Set font_width, font_height and font_base_line from this.font
   **/
  private void setFontInfo () {
    FontMetrics fm = getFontMetrics (getFont ());
      //      System.out.println ("fm:" + fm.toString ());

    int [] font_widths = fm.getWidths ();

    font_width = -1;
    // get the maximum character width

    for ( int i = 0 ; i < font_widths.length ; ++i ) {
      if (font_widths[i] > font_width) {
        font_width = font_widths[i];
      }
    }

//      font_width = font_width * 10 / 9;

//      System.out.println ("font width: " + font_width);

    font_height = fm.getHeight ();
    font_base_line = fm.getMaxAscent ();
  }

  /**
   *  Return the owning component that was passed to the constructor.
   **/
  private FeatureDisplayOwner getOwningComponent () {
    return owning_component;
  }

  /**
   *  Return the EntryGroup object that this FeatureDisplay is displaying.
   **/
  public EntryGroup getEntryGroup () {
    return entry_group;
  }

  /**
   *  Return an object that implements the GotoEventSource interface and which
   *  is relative to the sequence that this DisplayComponent is displaying.
   **/
  public GotoEventSource getGotoEventSource () {
    return goto_event_source;
  }

  /**
   *  Walk up through parents of this Component and return the Frame at the
   *  top.
   **/
  public Frame getParentFrame () {
    Component parent = this.getParent ();

    while (parent != null && !(parent instanceof Frame)) {
      parent = parent.getParent();
    }

    return (Frame) parent;
  }

  /**
   *  Update the value of scale_value to reflect the current value of
   *  scale_factor.
   **/
  private void setScaleValue () {
    final int scale_factor = getScaleFactor ();

    if (scale_factor > 0) {
      scale_value = (float) (1 / Math.pow (3, scale_factor - 1));
    } else {
      scale_value = getFontWidth ();
    }
  }

  /**
   *  Scroll and scale the display so that the given first base is at the left
   *  edge of the screen and the given last base is at the right edge.
   **/
  public void setFirstAndLastBase (final int first, final int last) {
    final int canvas_width = canvas.getSize ().width;

    first_visible_base = first;

    setScaleValue (1.0F * canvas_width / (last - first + 1));
  }

  /**
   *  Scroll the display so that the given first base is at the left edge of
   *  the screen.
   **/
  public void setFirstBase (int base_position) {
    if (base_position > getSequenceLength ()) {
      base_position = getSequenceLength ();
    }

    if (base_position < 1) {
      base_position = 1;
    }
    
    setFirstVisibleForwardBase (base_position);
  }

  /**
   *  Set the scale value to use for this FeatureDisplay.  Note that it is
   *  better to call setScaleFactor () which will also set the scale value.
   **/
  private void setScaleValue (final float scale_value) {
    if (scale_value <= 0) {
      throw new Error ("scale value must be positive");
    }
    this.scale_value = scale_value;

    if (scale_value > 1) {
      scale_factor = 1;
    } else {
      scale_factor =
        (int) Math.round (Math.log (1/scale_value) /  Math.log (3)) + 1;
    }


    scale_changer.setValue (scale_factor);

    fixScrollbar ();
    updateVisibleFeatureVector ();
    repaintCanvas ();
  }

  /**
   *  Make a new Range and throw a Error is an OutOfRangeException occurs.
   **/
  private Range newRange (final int start, final int end) {
    try {
      return new Range (start, end);
    } catch (OutOfRangeException e) {
      throw new Error ("internal error - unexpected exception: " + e);
    }
  }

  /**
   *  This is the EntryGroup that this component is displaying (as set by the
   *  constructor).
   **/
  private EntryGroup entry_group;

  /**
   *  Contains the canvas and it's scrollbar.
   **/
  private Panel map_panel = null;

  /**
   *  The drawing area for this component.
   **/
  private Canvas canvas = null;

  /**
   *  The Scrollbar for this FeatureDisplay object.  We create the scrollbar
   *  as part of this object rather than in the EntryEdit component because we
   *  may need to change the parameters of the scrollbar later.
   **/
  private Scrollbar scrollbar = null;

  /**
   *  A scroll bar for changing the viewing scale.
   **/
  private Scrollbar scale_changer = null;

  /**
   *  Used to colour one of the frames.
   **/
  private Color light_grey = new Color (220, 220, 220);

  /**
   *  Used to colour one of the frames.
   **/
  private Color lighter_grey = new Color (240, 240, 240);

  /**
   *  Used to colour one of the frames.
   **/
  private Color not_so_light_grey = new Color (200, 200, 200);

  /**
   *  This is a reference to the component that created this FeatureDisplay.
   *  We need this reference so that we can call owning_component.packme (),
   *  owning_component.getEntryGroup () etc.
   *
   **/
  private FeatureDisplayOwner owning_component;

  /**
   *  This is a reference to the Selection object that was passed to the
   *  constructor.
   **/
  private Selection selection;

  /**
   *  This is a reference to the GotoEventSource object that was passed to the
   *  constructor.
   **/
  private GotoEventSource goto_event_source;

  /**
   *  This Vector containing the references of those features that are
   *  currently visible.
   **/
  private FeatureVector visible_features = new FeatureVector ();

  /**
   *  A vector of those objects listening for adjustment events.
   **/
  final private Vector adjustment_listener_list = new Vector ();

  /**
   *  The height of the font used in this component.
   **/
  private int font_height;

  /**
   *  The (maximum) width of the font used in this component.
   **/
  private int font_width;

  /**
   *  The base line of the font used in this component.
   **/
  private int font_base_line;

  /**
   *  Off screen image used for double buffering when drawing the canvas.
   **/
  private Image offscreen;

  /**
   *  The index of the first base that we are displaying.
   **/
  private int first_visible_base = 1;

  /**
   *  See getScaleFactor ().
   **/
  private int scale_factor = 3;

  /**
   *  See getScaleValue ().
   **/
  private float scale_value = 1;

  /**
   *  This variable is true if labels should be shown.
   **/
  private boolean show_labels = true;

  /**
   *  This variable is true if stop codons should be shown.
   **/
  private boolean show_stop_codons = true;

  /**
   *  This variable is true if each base should be drawn in a different colour
   *  at scale feature 1.
   **/
  private boolean show_base_colours = false;

  /**
   *  All features (not just CDS features) will be drawn on the frame lines if
   *  and only if this variable is true.  See setFrameFeaturesFlag () and
   *  getFrameFeaturesFlag ().
   **/
  private boolean frame_features_flag = false;

  /**
   *  The position(s) of the last mouse click on the dna line.  The
   *  MarkerRange contains a reference to the appropriate Strand and contains
   *  the base positions.  See getMarkerRangeFromPosition () to understand why
   *  one click can give multiple bases.
   **/
  private MarkerRange click_range = null;

  /**
   *  The last (FeatureSegment) Marker that the user clicked on.  This is used
   *  for dragging the ends of segments.
   **/
  private Marker click_segment_marker = null;

  /**
   *  This is true if click_segment_marker is the Marker at the start of
   *  segment false otherwise.  The value is only useful if
   *  click_segment_marker is set.
   **/
  private boolean click_segment_marker_is_start_marker = false;

  /**
   *  When a FeatureSegment Marker drag starts, this is set to the Marker at
   *  the other end of the segment.  This is used to check that the drag has
   *  not move the Marker too far (past the end of the segment).
   **/
  private Marker other_end_of_segment_marker = null;

  /**
   *  Features with a /score qualifier less than this value will not be shown.
   **/
  private int current_min_score = 0;

  /**
   *  Features with a /score qualifier greater than this value will not be shown.
   **/
  private int current_max_score = 100;
}
