/* Algorithm.java
 *
 * created: Tue Dec 15 1998
 *
 * This file is part of Artemis
 * 
 * Copyright (C) 1998,1999,2000  Genome Research Limited
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /nfs/disk222/yeastpub/Repository/powmap/diana/plot/Algorithm.java,v 1.7 2000/01/24 17:04:21 kmr Exp $
 */

package diana.plot;

/**
 *  This class represents an algorithm that can be plotted.
 *
 *  @author Kim Rutherford
 *  @version $Id: Algorithm.java,v 1.7 2000/01/24 17:04:21 kmr Exp $
 **/

public abstract class Algorithm {
  /**
   *  Create a new Algorithm object.
   *  @param strand The strand to do the calculation on.
   *  @param algorithm_name A String used to identify this algorithm to the
   *    user.
   **/
  public Algorithm (String algorithm_name) {
    this.algorithm_name = algorithm_name;
  }

  /**
   *  Return the name of this algorithm.
   **/
  public String getAlgorithmName () {
     return algorithm_name;
  }

  /**
   *  Return the default or optimal window size.
   *  @return null is returned if this algorithm doesn't have optimal window
   *    size.
   **/
  public Integer getDefaultWindowSize () {
    return null;
  } 

  /**
   *  Return the default maximum window size for this algorithm.
   *  @return null is returned if this algorithm doesn't have maximum window
   *    size.
   **/
  public Integer getDefaultMaxWindowSize () {
    return null;
  } 

  /**
   *  Return the default minimum window size for this algorithm.
   *  @return null is returned if this algorithm doesn't have minimum window
   *    size.
   **/
  public Integer getDefaultMinWindowSize () {
    return null;
  } 

  /**
   *  Return the default or optimal step size.
   *  @return null is returned if this algorithm doesn't have optimal step
   *    size.
   **/
  public Integer getDefaultStepSize (int window_size) {
    return null;
  } 

  /**
   *  Return the maximum value of this algorithm.    Subclasses of this class
   *  should override getMaximumInternal () rather than this method.
   *  @return null is returned if this algorithm doesn't have a fixed maximum
   *    or if maxAndMinDisabled () is true.
   **/
  final public Float getMaximum () {
    if (scalingFlag ()) {
      return null;
    } else {
      return getMaximumInternal ();
    }
  } 

  /**
   *  Return the maximum value of this algorithm.
   *  @return null is returned if this algorithm doesn't have a fixed maximum.
   **/
  protected Float getMaximumInternal () {
    return null;
  } 

  /**
   *  Return the minimum value of this algorithm.  Subclasses of this class
   *  should override getMinimumInternal () rather than this method.
   *  @return null is returned if this algorithm doesn't have a fixed minimum
   *    or if maxAndMinDisabled () is true.
   **/
  final public Float getMinimum () {
    if (scalingFlag ()) {
      return null;
    } else {
      return getMinimumInternal ();
    }
  } 

  /**
   *  Return the minimum value of this algorithm.
   *  @return null is returned if this algorithm doesn't have a fixed minimum.
   **/
  protected Float getMinimumInternal () {
    return null;
  } 

  /**
   *  Return the average value of function over the whole strand.
   *  @return null is returned if this algorithm doesn't have an average or if
   *    the average can't be calculated.
   **/
  public Float getAverage () {
    return null;
  }

  /**
   *  Force getMinimum () and getMaximum () to return null if and only if
   *  enable_scaling is true.
   **/
  public void setScalingFlag (final boolean enable_scaling) {
    max_min_disabled = enable_scaling;
  }

  /**
   *  Return true if and only if getMaximum () and getMinimum () should return
   *  null.
   **/
  public boolean scalingFlag () {
    return max_min_disabled;
  }
  
  /**
   *  The name of this algorithm, as passed to the constructor.
   **/
  private String algorithm_name;


  /**
   *  Set by disableMaxAndMin () and enableMaxAndMin ().
   **/
  private boolean max_min_disabled = false;
}
